@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 2002, 2003 Free Software Foundation, Inc.
+ * Copyright 2002, 2003, 2004, 2005, 2006 Free Software Foundation, Inc.
  * Written by Paolo Bonzini.  Ideas based on Mike Haertel's malloc.
  *
  * This file is part of GNU Smalltalk.
@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -24,7 +46,7 @@
  * 
  * You should have received a copy of the GNU General Public License along with
  * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
  ***********************************************************************/
 
 #include "gstpriv.h"
@@ -32,15 +54,28 @@
 /* Define to provide a log of memory allocations on stderr.  */
 /* #define LOG_MEMORY_OPERATIONS */
 
+/* MinGW32 and MacOS X have problems redefining malloc and friends.  In this
+   case, we use a completely separate area than the one used by the system
+   malloc and we don't cause trouble.
 
-#define	SMALL2FREE(B, N)	((heap_freeobj*)(&(B)->vSmall.data[(N)*(B)->size]))
+   For MacOS X, these problems only show up when using readline: readline
+   will use the system malloc because of Mach-O's unique two-level namespaces
+   "feature", and as you can imagine we have some problems when we try to use
+   *our* routine to reallocate memory that readline has allocated.  Even
+   if we did not try reallocating readline memory, we would not have a way
+   to free it.  */
+#if defined __MSVCRT__ || defined __MACH__
+#define NO_MALLOC_OVERRIDE
+#endif
+
+#define	SMALL2FREE(B, N)	((heap_freeobj*)(((char *)(B)->vSmall.data) + (N)*(B)->size))
 
 #define	MEM2BLOCK(M)		((heap_block*)(((intptr_t)(M)) & -pagesize))
 #define	MEM2FREE(M)		((heap_freeobj*)(M))
 
 #define	BLOCKEND(B)		((heap_block*)(((unsigned char*)(B)) + (B)->size))
 
-#define	MAX_SMALL_OBJ_SIZE	8192
+#define	MAX_SMALL_OBJ_SIZE	16384
 #define	IS_SMALL_SIZE(S)	((S) <= max_small_object_size)
 
 #define	MEMALIGN		8
@@ -62,11 +97,6 @@
 
 static void init_heap (heap_data *h, size_t heap_allocation_size, size_t heap_limit);
 
-/* Oh my God, how broken things sometimes are!  */
-#if defined small
-# undef small
-#endif
-
 #define vSmall var.small
 #define vLarge var.large
 #define vFree  var.free
@@ -92,11 +122,14 @@ static void log_memory (char *op, int n, PTR block);
      bc | uniq | sed '$!s/$/,/' | fmt -60
 
    for 32-bit machines, and similarly with 4064 instead of
-   4072 on 64-bit machines.  8 and 16 were added manually.  */
+   4072 for 64-bit machines.  8 and 16 were added manually.  */
 
 static unsigned short freelist_size[NUM_FREELISTS + 1] = {
   8, 16, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352,
-  384, 448, 480, 576, 672, 800, 992, 1344, 2016, 4064, 0
+  384, 448, 480, 576, 672, 800, 992, 1344, 2016,
+  4096 - SMALL_OBJ_HEADER_SIZE,
+  8192 - SMALL_OBJ_HEADER_SIZE,
+  16384 - SMALL_OBJ_HEADER_SIZE, 0
 };
 
 static unsigned short sztable[MAX_SMALL_OBJ_SIZE + 1];
@@ -128,7 +161,7 @@ init_heap (heap_data *h, size_t heap_allocation_size, size_t heap_limit)
 
       /* Use the preinitialized freelist table to initialize
          the sztable.  */
-      for (sz = i = 0; freelist_size[i] > 0; i++)
+      for (sz = i = 0; freelist_size[i] > 0 && freelist_size[i] < pagesize; i++)
         for (; sz <= freelist_size[i]; sz++)
           sztable[sz] = i;
       max_small_object_size = sz - 1;
@@ -137,8 +170,9 @@ init_heap (heap_data *h, size_t heap_allocation_size, size_t heap_limit)
   for (i = 0; freelist_size[i] > 0; i++)
     h->freelist[i] = NULL;
 
-  h->heap_allocation_size = heap_allocation_size ? pagesize :
-    ROUNDUPPAGESIZE (heap_allocation_size);
+  h->heap_allocation_size = (heap_allocation_size
+			     ? ROUNDUPPAGESIZE (heap_allocation_size)
+			     : MMAP_THRESHOLD);
   h->heap_limit = heap_limit;
   h->mmap_count = 0;
   h->heap_total = 0;
@@ -185,25 +219,26 @@ rerun:
 	      goto nospace;
 	    }
 
+#ifndef OPTIMIZE
           if (((intptr_t) blk) & (pagesize - 1))
 	    abort ();
+#endif
 
 	  blk->vSmall.nfree = *mptr;
 	  *mptr = blk;
 	}
 
-      if (!blk->vSmall.free)
-	abort ();
-
-      if (!blk->vSmall.avail)
-	abort ();
-
       /* Unlink free one and return it */
       mem = blk->vSmall.free;
 
+#ifndef OPTIMIZE
+      if (!blk->vSmall.free || !blk->vSmall.avail)
+	abort ();
+
       if (((intptr_t) mem <= (intptr_t) blk) ||
 	  ((intptr_t) mem >= (intptr_t) blk + pagesize))
 	abort ();
+#endif
 
       blk->vSmall.free = mem->next;
 
@@ -226,8 +261,10 @@ rerun:
       mem = (heap_freeobj *) blk->vLarge.data;
     }
 
+#ifndef OPTIMIZE
   if (OBJECT_SIZE (mem) < sz)
     abort ();
+#endif
 
   if (h->after_allocating)
     h->after_allocating (h, blk, sz);
@@ -337,8 +374,10 @@ _gst_mem_free (heap_data *h, PTR mem)
          it to freelist.  */
       if (++info->vSmall.avail == 1)
 	{
+#ifndef OPTIMIZE
 	  if ( ((intptr_t) info) & (pagesize - 1))
 	    abort ();
+#endif
 
 	  info->vSmall.nfree = h->freelist[lnr];
 	  h->freelist[lnr] = info;
@@ -347,17 +386,20 @@ _gst_mem_free (heap_data *h, PTR mem)
       obj->next = info->vSmall.free;
       info->vSmall.free = obj;
 
+#ifndef OPTIMIZE
       if ((intptr_t) obj < (intptr_t) info ||
 	  (intptr_t) obj >= (intptr_t) info + pagesize ||
 	  (intptr_t) obj == (intptr_t) (obj->next))
 	abort ();
 
+      if (info->vSmall.avail > info->vSmall.nr)
+	abort ();
+#endif
+
       /* If we free all sub-blocks, free the block */
-      if (info->vSmall.avail >= info->vSmall.nr)
+      if (info->vSmall.avail == info->vSmall.nr)
 	{
 	  heap_block **finfo = &h->freelist[lnr];
-          if (info->vSmall.avail > info->vSmall.nr)
-	    abort ();
 
 	  for (;;)
 	    {
@@ -369,8 +411,10 @@ _gst_mem_free (heap_data *h, PTR mem)
 		  break;
 		}
 	      finfo = &(*finfo)->vSmall.nfree;
+#ifndef OPTIMIZE
 	      if (!*finfo)
 		abort ();
+#endif
 	    }
 	}
     }
@@ -450,8 +494,10 @@ heap_primitive_alloc (heap_data *h, size_t sz)
       && h->heap_total + sz > h->heap_limit)
     return (NULL);
 
+#ifndef OPTIMIZE
   if (sz & (pagesize - 1))
     abort ();
+#endif
 
   if (sz > MMAP_THRESHOLD)
     {
@@ -478,9 +524,14 @@ heap_primitive_alloc (heap_data *h, size_t sz)
   for (pptr = &heap_prim_freelist; (ptr = *pptr); pptr = &(ptr->vFree.next))
     {
       h->probes++;
+#ifndef OPTIMIZE
       if (((intptr_t) ptr) & (pagesize - 1))
 	abort ();
 
+      if (ptr->size & (pagesize - 1))
+	abort ();
+#endif
+
       /* First fit */
       if (sz <= ptr->size)
 	{
@@ -510,9 +561,6 @@ heap_primitive_alloc (heap_data *h, size_t sz)
 
 	  return (ptr);
 	}
-
-      if (ptr->size & (pagesize - 1))
-	abort ();
     }
 
   /* Nothing found on free list */
@@ -524,8 +572,10 @@ heap_primitive_alloc (heap_data *h, size_t sz)
 static void
 heap_primitive_free (heap_data *h, heap_block *mem)
 {
+#ifndef OPTIMIZE
   if (mem->size & (pagesize - 1))
     abort ();
+#endif
 
   if (h->before_prim_freeing)
     h->before_prim_freeing (h, mem, mem->size);
@@ -546,9 +596,14 @@ heap_add_to_free_list (heap_data *h, heap_block *mem)
   heap_block *lptr;
   heap_block *nptr;
 
+#ifndef OPTIMIZE
   if (((intptr_t) mem) & (pagesize - 1))
     abort ();
 
+  if (mem->size & (pagesize - 1))
+    abort ();
+#endif
+
   if (mem < heap_prim_freelist || heap_prim_freelist == 0)
     {
       /* If this block is directly before the first block on the
@@ -570,8 +625,10 @@ heap_add_to_free_list (heap_data *h, heap_block *mem)
   lptr = heap_prim_freelist;
   while (lptr->vFree.next != 0)
     {
+#ifndef OPTIMIZE
       if (lptr->size & (pagesize - 1))
 	abort ();
+#endif
 
       nptr = lptr->vFree.next;
       if (mem > lptr && mem < nptr)
@@ -626,8 +683,10 @@ static void
 heap_system_alloc (heap_data *h, size_t sz)
 {
   heap_block * mem;
+#ifndef OPTIMIZE
   if (sz & (pagesize - 1))
     abort ();
+#endif
 
   mem = (heap_block *) morecore (sz);
   mem->mmap_block = 0;
@@ -641,13 +700,25 @@ PTR
 morecore (size_t size)
 {
   heap just_allocated_heap = NULL;
-  static heap current_extra_heap = NULL;
+
+  /* _gst_heap_sbrk is actually the same as sbrk as long as
+     current_heap is NULL.  But we cannot do that unless we
+     can replace malloc (which we cannot do on MacOS X, see above).  */
+  static heap current_heap = NULL;
+
+#ifdef NO_MALLOC_OVERRIDE
+  if (current_heap == NULL)
+    {
+      just_allocated_heap = _gst_heap_create (NULL, MMAP_AREA_SIZE);
+      if (!just_allocated_heap)
+	return (NULL);
+      current_heap = just_allocated_heap;
+    }
+#endif
 
   for (;;)
     {
-      /* _gst_heap_sbrk is actually the same as sbrk as long as
-         current_extra_heap is NULL.  */
-      char *ptr = _gst_heap_sbrk (current_extra_heap, size);
+      char *ptr = _gst_heap_sbrk (current_heap, size);
 
       if (ptr != (PTR) -1)
 	{
@@ -655,8 +726,8 @@ morecore (size_t size)
             {
 	      /* Oops, we have to align to a page.  */
 	      int missed = pagesize - ((intptr_t) ptr & (pagesize - 1));
-	      _gst_heap_sbrk (current_extra_heap, -size + missed);
-	      ptr = _gst_heap_sbrk (current_extra_heap, size);
+	      _gst_heap_sbrk (current_heap, -size + missed);
+	      ptr = _gst_heap_sbrk (current_heap, size);
             }
 
           if (ptr != (PTR) -1)
@@ -670,22 +741,16 @@ morecore (size_t size)
       if (just_allocated_heap)
 	return (NULL);
 
-      just_allocated_heap = _gst_heap_create (MMAP_AREA_SIZE);
+      just_allocated_heap = _gst_heap_create (NULL, MMAP_AREA_SIZE);
       if (!just_allocated_heap)
 	return (NULL);
 
-      current_extra_heap = just_allocated_heap;
+      current_heap = just_allocated_heap;
     }
 }
 
 
-/* MinGW32 has problems redefining malloc and friends.  It's not
-   a problem: since sbrk is not present there, we end up using a
-   completely separate area than the one used by the system malloc
-   and we don't cause trouble.  */
-
-#ifndef __MSVCRT__
-
+#ifndef NO_MALLOC_OVERRIDE
 static heap_data default_heap;
 static int default_heap_init;
 
@@ -760,7 +825,7 @@ free (PTR block)
 #endif
   _gst_mem_free(&default_heap, block);
 }
-#endif /* !__MSVCRT__ */
+#endif /* !NO_MALLOC_OVERRIDE */
 
 char *
 xstrdup (const char *str)
