@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -25,12 +47,13 @@
  * 
  * You should have received a copy of the GNU General Public License along with
  * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
  *
  ***********************************************************************/
 
 
 #include "gstpriv.h"
+#include "re.h"
 
 #include "../libffi/include/ffi.h"
 #include <ltdl.h>
@@ -40,10 +63,10 @@ typedef enum
   CDATA_UNKNOWN,		/* when there is no type a priori */
   CDATA_CHAR,
   CDATA_STRING,
-  CDATA_STRINGOUT,		/* for things that modify string params */
+  CDATA_STRING_OUT,		/* for things that modify string params */
   CDATA_SYMBOL,
   CDATA_BYTEARRAY,
-  CDATA_BYTEARRAYOUT,
+  CDATA_BYTEARRAY_OUT,
   CDATA_BOOLEAN,
   CDATA_INT,
   CDATA_UINT,
@@ -66,26 +89,14 @@ typedef enum
   CDATA_OOP,			/* no conversion to-from C (OOP) */
   CDATA_SELF,			/* pass self as the corresponding
 				   argument */
-  CDATA_SELF_OOP		/* pass self as an OOP */
+  CDATA_SELF_OOP,		/* pass self as an OOP */
+  CDATA_WCHAR,
+  CDATA_WSTRING,
+  CDATA_WSTRING_OUT,
+  CDATA_SYMBOL_OUT
 }
 cdata_type;
 
-typedef struct gst_cfunc_descriptor
-{
-  OBJ_HEADER;
-  OOP cFunction;		/* gst_cobject whose C value is func
-				   addr */
-  OOP cFunctionName;		/* Name of C function in mapping table */
-  OOP returnType;		/* Smalltalk return type */
-  OOP numFixedArgs;		/* number of real arguments passed from
-				   smalltalk (excluding "self" parameters
-  				   which are synthetically added when
-				   calling the C function).  */
-  OOP argTypes[1];		/* variable length, really numFixedArgs 
-				   long */
-}
- *gst_cfunc_descriptor;
-
 typedef struct symbol_type_map
 {
   OOP *symbol;
@@ -198,6 +209,8 @@ static int get_errno (void);
 /* Encapsulate binary incompatibilities between various C libraries.  */
 static int my_stat (const char *name,
 		    gst_stat * out);
+static int my_lstat (const char *name,
+		     gst_stat * out);
 static int my_putenv (const char *str);
 static int my_chdir (const char *str);
 static DIR *my_opendir (const char *str);
@@ -218,10 +231,10 @@ static const char *c_type_name[] = {
   "void?",			/* CDATA_UNKNOWN */
   "char",			/* CDATA_CHAR */
   "char *",			/* CDATA_STRING */
-  "char *",			/* CDATA_STRINGOUT */
+  "char *",			/* CDATA_STRING_OUT */
   "char *",			/* CDATA_SYMBOL */
   "char *",			/* CDATA_BYTEARRAY */
-  "char *",			/* CDATA_BYTEARRAYOUT */
+  "char *",			/* CDATA_BYTEARRAY_OUT */
   "int",			/* CDATA_BOOLEAN */
   "int",			/* CDATA_INT */
   "unsigned int",		/* CDATA_UINT */
@@ -238,6 +251,9 @@ static const char *c_type_name[] = {
   "OOP",			/* CDATA_OOP */
   "void?",			/* CDATA_SELF */
   "OOP",			/* CDATA_SELF_OOP */
+  "wchar_t",			/* CDATA_WCHAR */
+  "wchar_t *",			/* CDATA_WSTRING */
+  "wchar_t *",			/* CDATA_WSTRING_OUT */
 };
 
 /* A map between symbols and the cdata_type enum.  */
@@ -245,10 +261,11 @@ static const symbol_type_map type_map[] = {
   {&_gst_unknown_symbol, CDATA_UNKNOWN},
   {&_gst_char_symbol, CDATA_CHAR},
   {&_gst_string_symbol, CDATA_STRING},
-  {&_gst_string_out_symbol, CDATA_STRINGOUT},
+  {&_gst_string_out_symbol, CDATA_STRING_OUT},
   {&_gst_symbol_symbol, CDATA_SYMBOL},
+  {&_gst_symbol_out_symbol, CDATA_SYMBOL_OUT},
   {&_gst_byte_array_symbol, CDATA_BYTEARRAY},
-  {&_gst_byte_array_out_symbol, CDATA_BYTEARRAYOUT},
+  {&_gst_byte_array_out_symbol, CDATA_BYTEARRAY_OUT},
   {&_gst_boolean_symbol, CDATA_BOOLEAN},
   {&_gst_int_symbol, CDATA_INT},
   {&_gst_uint_symbol, CDATA_UINT},
@@ -265,6 +282,9 @@ static const symbol_type_map type_map[] = {
   {&_gst_smalltalk_symbol, CDATA_OOP},
   {&_gst_self_symbol, CDATA_SELF},
   {&_gst_self_smalltalk_symbol, CDATA_SELF_OOP},
+  {&_gst_wchar_symbol, CDATA_WCHAR},
+  {&_gst_wstring_symbol, CDATA_WSTRING},
+  {&_gst_wstring_out_symbol, CDATA_WSTRING_OUT},
   {NULL, CDATA_UNKNOWN}
 };
 
@@ -305,16 +325,37 @@ my_stat (const char *name,
       errno = 0;
       out->st_mode = statOut.st_mode;
       out->st_size = statOut.st_size;
-      out->st_aTime =
-	_gst_adjust_time_zone (statOut.st_atime) - 86400 * 10959;
-      out->st_mTime =
-	_gst_adjust_time_zone (statOut.st_mtime) - 86400 * 10959;
-      out->st_cTime =
-	_gst_adjust_time_zone (statOut.st_ctime) - 86400 * 10959;
+      out->st_aTime = _gst_adjust_time_zone (statOut.st_atime) - 86400 * 10957;
+      out->st_mTime = _gst_adjust_time_zone (statOut.st_mtime) - 86400 * 10957;
+      out->st_cTime = _gst_adjust_time_zone (statOut.st_ctime) - 86400 * 10957;
     }
   return (result);
 }
 
+#ifdef HAVE_LSTAT
+int
+my_lstat (const char *name,
+	 gst_stat * out)
+{
+  int result;
+  static struct stat statOut;
+
+  result = lstat (name, &statOut);
+  if (!result)
+    {
+      errno = 0;
+      out->st_mode = statOut.st_mode;
+      out->st_size = statOut.st_size;
+      out->st_aTime = _gst_adjust_time_zone (statOut.st_atime) - 86400 * 10957;
+      out->st_mTime = _gst_adjust_time_zone (statOut.st_mtime) - 86400 * 10957;
+      out->st_cTime = _gst_adjust_time_zone (statOut.st_ctime) - 86400 * 10957;
+    }
+  return (result);
+}
+#else
+#define my_lstat my_stat
+#endif
+
 int
 my_putenv (const char *str)
 {
@@ -395,7 +436,7 @@ get_argc (void)
 const char *
 get_argv (int n)
 {
-  return (n <= _gst_smalltalk_passed_argc
+  return (n >= 1 && n <= _gst_smalltalk_passed_argc
 	  ? _gst_smalltalk_passed_argv[n - 1]
 	  : NULL);
 }
@@ -405,14 +446,14 @@ dld_open (const char *filename)
 {
 #ifdef ENABLE_DLD
   lt_dlhandle handle;
-  void (*initModule) ();
+  void (*initModule) (struct VMProxy *);
 
   handle = lt_dlopenext (filename);
   if (handle)
     {
       initModule = lt_dlsym (handle, "gst_initModule");
       if (initModule)
-	initModule (&gst_interpreter_proxy);
+	initModule (_gst_get_vmproxy ());
     }
 
   return (handle);
@@ -462,9 +503,9 @@ _gst_init_cfuncs (void)
   _gst_define_cfunc ("errno", get_errno);
   _gst_define_cfunc ("strerror", strerror);
   _gst_define_cfunc ("stat", my_stat);
-#if 0
   _gst_define_cfunc ("lstat", my_lstat);
-#endif
+  _gst_define_cfunc ("utime", _gst_set_file_access_times);
+  _gst_define_cfunc ("chmod", chmod);
 
   _gst_define_cfunc ("opendir", my_opendir);
   _gst_define_cfunc ("closedir", closedir);
@@ -472,11 +513,13 @@ _gst_init_cfuncs (void)
   _gst_define_cfunc ("rewinddir", rewinddir);
   _gst_define_cfunc ("extractDirentName", extract_dirent_name);
 
+  _gst_define_cfunc ("symlink", symlink);
   _gst_define_cfunc ("unlink", unlink);
   _gst_define_cfunc ("rename", rename);
   _gst_define_cfunc ("rmdir", rmdir);
   _gst_define_cfunc ("chdir", my_chdir);
   _gst_define_cfunc ("mkdir", mkdir);
+  _gst_define_cfunc ("mkdtemp", mkdtemp);
   _gst_define_cfunc ("getCurDirName", _gst_get_cur_dir_name);
 
   _gst_define_cfunc ("fileIsReadable", _gst_file_is_readable);
@@ -485,6 +528,11 @@ _gst_init_cfuncs (void)
 
   init_dld ();
 
+  /* regex routines */
+  _gst_define_cfunc ("reh_search", _gst_re_search);
+  _gst_define_cfunc ("reh_match", _gst_re_match);
+  _gst_define_cfunc ("reh_make_cacheable", _gst_re_make_cacheable);
+
   /* Non standard routines */
   _gst_define_cfunc ("marli", marli);
 }
@@ -564,6 +612,9 @@ _gst_invoke_croutine (OOP cFuncOOP,
   incPtr = INC_SAVE_POINTER ();
 
   desc = (gst_cfunc_descriptor) OOP_TO_OBJ (cFuncOOP);
+  if (IS_NIL (desc->cFunction))
+    return (NULL);
+
   c_func_cur = (cfunc_info *) COBJECT_VALUE (desc->cFunction);
   if (!c_func_cur)
     return (NULL);
@@ -606,7 +657,7 @@ _gst_invoke_croutine (OOP cFuncOOP,
   for (si = i = 0; i < fixedArgs; i++)
     {
       cType = TO_INT (desc->argTypes[i]);
-      if (cType >= CDATA_SELF)
+      if (cType == CDATA_SELF || cType == CDATA_SELF_OOP)
 	push_smalltalk_obj (receiver,
 			    cType == CDATA_SELF ? CDATA_UNKNOWN : CDATA_OOP);
       else if (cType != CDATA_VOID)
@@ -648,11 +699,15 @@ _gst_invoke_croutine (OOP cFuncOOP,
 	    SET_COBJECT_VALUE (arg->oop, arg->u.cObjectPtrVal.ptrVal);
 	    continue;
 
-          case CDATA_STRINGOUT:
+          case CDATA_WSTRING_OUT:
+	    _gst_set_oop_unicode_string (arg->oop, arg->u.ptrVal);
+	    break;
+
+          case CDATA_STRING_OUT:
 	    _gst_set_oopstring (arg->oop, arg->u.ptrVal);
 	    break;
 
-          case CDATA_BYTEARRAYOUT:
+          case CDATA_BYTEARRAY_OUT:
 	    _gst_set_oop_bytes (arg->oop, arg->u.ptrVal);
 	    break;
 
@@ -679,21 +734,29 @@ get_ffi_type (OOP returnTypeOOP)
     case CDATA_COBJECT:
     case CDATA_COBJECT_PTR:
     case CDATA_SYMBOL:
+    case CDATA_SYMBOL_OUT:
+    case CDATA_WSTRING:
+    case CDATA_WSTRING_OUT:
     case CDATA_STRING:
-    case CDATA_STRINGOUT:
+    case CDATA_STRING_OUT:
     case CDATA_BYTEARRAY:
-    case CDATA_BYTEARRAYOUT:
+    case CDATA_BYTEARRAY_OUT:
     default:
       return &ffi_type_pointer;
 
     case CDATA_LONG:
     case CDATA_ULONG:
-      return &ffi_type_slong;
+#if LONG_MAX == 2147483647
+      return &ffi_type_sint32;
+#else
+      return &ffi_type_sint64;
+#endif
 
     case CDATA_VOID:
     case CDATA_INT:
     case CDATA_UINT:
     case CDATA_CHAR:
+    case CDATA_WCHAR:
     case CDATA_BOOLEAN:
       return &ffi_type_sint;
 
@@ -726,23 +789,51 @@ push_smalltalk_obj (OOP oop,
   int i;
   cparam *cp = &c_func_cur->args[c_func_cur->arg_idx];
 
-  cp->oop = NULL;
   class = OOP_INT_CLASS (oop);
 
-  if (cType == CDATA_UNKNOWN)
+  switch (cType)
     {
+    case CDATA_UNKNOWN:
       cType =
 	(oop == _gst_true_oop || oop == _gst_false_oop) ? CDATA_BOOLEAN :
 	oop == _gst_nil_oop ? CDATA_COBJECT :
 	class == _gst_char_class ? CDATA_CHAR :
+	class == _gst_unicode_character_class ? CDATA_WCHAR :
 	class == _gst_byte_array_class ? CDATA_BYTEARRAY : 
 	is_a_kind_of (class, _gst_integer_class) ? CDATA_LONG :
 	is_a_kind_of (class, _gst_string_class) ? CDATA_STRING :
+	is_a_kind_of (class, _gst_unicode_string_class) ? CDATA_WSTRING :
 	is_a_kind_of (class, _gst_c_object_class) ? CDATA_COBJECT :
 	is_a_kind_of (class, _gst_float_class) ? CDATA_DOUBLE :
 	CDATA_OOP;
+      break;
+
+    case CDATA_VARIADIC:
+      if (class == _gst_array_class)
+	{
+	  c_func_cur->cacheVariadic = true;
+	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
+	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_UNKNOWN);
+	}
+      else
+        bad_type (class, cType);
+
+      return;
+
+    case CDATA_VARIADIC_OOP:
+      if (class == _gst_array_class)
+	{
+	  c_func_cur->cacheVariadic = true;
+	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
+	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_OOP);
+	}
+      else
+        bad_type (class, cType);
+
+      return;
     }
 
+  cp->oop = NULL;
   cp->cType = cType;
 
   if (cType == CDATA_OOP)
@@ -760,7 +851,11 @@ push_smalltalk_obj (OOP oop,
         case CDATA_LONG:
 	case CDATA_ULONG:
 	  cp->u.longVal = TO_C_LONG (oop);
-	  SET_TYPE (&ffi_type_slong);
+#if LONG_MAX == 2147483647
+          SET_TYPE (&ffi_type_sint32);
+#else
+          SET_TYPE (&ffi_type_sint64);
+#endif
 	  return;
 
         case CDATA_INT:
@@ -789,7 +884,11 @@ push_smalltalk_obj (OOP oop,
         case CDATA_LONG:
 	case CDATA_ULONG:
 	  cp->u.longVal = (oop == _gst_true_oop);
-	  SET_TYPE (&ffi_type_slong);
+#if LONG_MAX == 2147483647
+          SET_TYPE (&ffi_type_sint32);
+#else
+          SET_TYPE (&ffi_type_sint64);
+#endif
 	  return;
 
         case CDATA_INT:
@@ -802,7 +901,9 @@ push_smalltalk_obj (OOP oop,
 	}
     }
 
-  else if (class == _gst_char_class && cType == CDATA_CHAR)
+  else if ((class == _gst_char_class
+	    && (cType == CDATA_CHAR || cType == CDATA_WCHAR))
+           || (class == _gst_unicode_character_class && cType == CDATA_WCHAR))
     {
       cp->u.intVal = CHAR_OOP_VALUE (oop);
       SET_TYPE (&ffi_type_sint);
@@ -810,14 +911,14 @@ push_smalltalk_obj (OOP oop,
     }
 
   else if (((class == _gst_string_class || class == _gst_byte_array_class)
-            && (cType == CDATA_STRING || cType == CDATA_STRINGOUT
-	        || cType == CDATA_BYTEARRAY || cType == CDATA_BYTEARRAYOUT))
+            && (cType == CDATA_STRING || cType == CDATA_STRING_OUT
+	        || cType == CDATA_BYTEARRAY || cType == CDATA_BYTEARRAY_OUT))
            || (class == _gst_symbol_class
                && (cType == CDATA_SYMBOL || cType == CDATA_STRING)))
     {
       cp->oop = oop;
 
-      if (cp->cType == CDATA_BYTEARRAY || cp->cType == CDATA_BYTEARRAYOUT)
+      if (cp->cType == CDATA_BYTEARRAY || cp->cType == CDATA_BYTEARRAY_OUT)
 	cp->u.ptrVal = _gst_to_byte_array (oop);
       else
         cp->u.ptrVal = (gst_uchar *) _gst_to_cstring (oop);
@@ -827,6 +928,17 @@ push_smalltalk_obj (OOP oop,
       return;
     }
 
+  else if (class == _gst_unicode_string_class
+           && (cType == CDATA_WSTRING || cType == CDATA_WSTRING_OUT))
+    {
+      cp->oop = oop;
+      cp->u.ptrVal = (gst_uchar *) _gst_to_wide_cstring (oop);
+
+      c_func_cur->needPostprocessing = true;
+      SET_TYPE (&ffi_type_pointer);
+      return;
+    }
+
   else if (is_a_kind_of (class, _gst_float_class))
     {
       switch (cType)
@@ -876,9 +988,9 @@ push_smalltalk_obj (OOP oop,
 	{
 	case CDATA_COBJECT:
 	case CDATA_BYTEARRAY:
-	case CDATA_BYTEARRAYOUT:
+	case CDATA_BYTEARRAY_OUT:
 	case CDATA_STRING:
-	case CDATA_STRINGOUT:
+	case CDATA_STRING_OUT:
 	case CDATA_SYMBOL:
 	  cp->u.ptrVal = NULL;
 	  SET_TYPE (&ffi_type_pointer);
@@ -886,25 +998,6 @@ push_smalltalk_obj (OOP oop,
 	}
     }
 
-  else if (class == _gst_array_class)
-    {
-      switch (cType)
-	{
-	case CDATA_VARIADIC:
-	  c_func_cur->cacheVariadic = true;
-	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
-	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_UNKNOWN);
-
-	  return;
-
-        case CDATA_VARIADIC_OOP:
-	  c_func_cur->cacheVariadic = true;
-	  for (i = 1; i <= NUM_WORDS (OOP_TO_OBJ (oop)); i++)
-	    push_smalltalk_obj (ARRAY_AT (oop, i), CDATA_OOP);
-
-	  return;
-	}
-    }
 
   bad_type (class, cType);
 }
@@ -930,6 +1023,10 @@ c_to_smalltalk (cparam *result, OOP returnTypeOOP)
       resultOOP = CHAR_OOP_AT ((gst_uchar) result->u.intVal);
       break;
 
+    case CDATA_WCHAR:
+      resultOOP = char_new ((wchar_t) result->u.intVal);
+      break;
+
     case CDATA_BOOLEAN:
       resultOOP = result->u.intVal ? _gst_true_oop : _gst_false_oop;
       break;
@@ -951,36 +1048,45 @@ c_to_smalltalk (cparam *result, OOP returnTypeOOP)
       break;
 
     case CDATA_STRING:
-    case CDATA_STRINGOUT:
+    case CDATA_STRING_OUT:
+    case CDATA_WSTRING:
+    case CDATA_WSTRING_OUT:
     case CDATA_SYMBOL:
+    case CDATA_SYMBOL_OUT:
     case CDATA_COBJECT:
     case CDATA_OOP:
       if (!result->u.ptrVal)
-	{
-	  resultOOP = _gst_nil_oop;
-	  break;
-	}
-      if (returnType == CDATA_SYMBOL)
+	resultOOP = _gst_nil_oop;
+      else if (returnType == CDATA_OOP)
+	resultOOP = (OOP) result->u.ptrVal;
+
+      else if (returnType == CDATA_SYMBOL || returnType == CDATA_SYMBOL_OUT)
 	{
 	  resultOOP = _gst_intern_string ((char *) result->u.ptrVal);
-	  break;
+	  if (returnType == CDATA_SYMBOL_OUT)
+	    xfree (result->u.ptrVal);
 	}
-      if (returnType == CDATA_COBJECT)
+      else if (returnType == CDATA_COBJECT)
 	{
 	  if (IS_INT (returnTypeOOP))
 	    resultOOP = COBJECT_NEW (result->u.ptrVal);
 	  else
 	    resultOOP = _gst_c_object_new_typed (result->u.ptrVal, returnTypeOOP);
-	  break;
 	}
-      if (returnType == CDATA_OOP)
+      else if (returnType == CDATA_STRING || returnType == CDATA_STRING_OUT)
 	{
-	  resultOOP = (OOP) result->u.ptrVal;
-	  break;
+	  resultOOP = _gst_string_new ((char *) result->u.ptrVal);
+	  if (returnType == CDATA_STRING_OUT)
+	    xfree (result->u.ptrVal);
 	}
-      resultOOP = _gst_string_new ((char *) result->u.ptrVal);
-      if (returnType == CDATA_STRINGOUT)
-	xfree (result->u.ptrVal);
+      else if (returnType == CDATA_WSTRING || returnType == CDATA_WSTRING_OUT)
+	{
+	  resultOOP = _gst_unicode_string_new ((wchar_t *) result->u.ptrVal);
+	  if (returnType == CDATA_WSTRING_OUT)
+	    xfree (result->u.ptrVal);
+	}
+      else
+	abort ();
       break;
 
     case CDATA_DOUBLE:
@@ -1110,17 +1216,3 @@ _gst_set_errno(int errnum)
   _gst_errno = errnum;
 #endif
 }
-
-void
-_gst_restore_cfunc_descriptor (OOP cFuncDescOOP)
-{
-  gst_cfunc_descriptor desc;
-  cfunc_info *cfi;
-  char *funcName;
-
-  desc = (gst_cfunc_descriptor) OOP_TO_OBJ (cFuncDescOOP);
-  funcName = (char *) _gst_to_cstring (desc->cFunctionName);
-  cfi = lookup_function (funcName);
-  xfree (funcName);
-  SET_COBJECT_VALUE (desc->cFunction, cfi);
-}
