@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -42,7 +64,8 @@ typedef struct class_definition
   OOP *classVar;
   OOP *superClassPtr;
   intptr_t instanceSpec;
-  char numFixedFields;
+  mst_Boolean reloadAddress;
+  int numFixedFields;
   const char *name;
   const char *instVarNames;
   const char *classVarNames;
@@ -78,6 +101,7 @@ OOP _gst_compiled_block_class = NULL;
 OOP _gst_compiled_code_class = NULL;
 OOP _gst_compiled_method_class = NULL;
 OOP _gst_context_part_class = NULL;
+OOP _gst_continuation_class = NULL;
 OOP _gst_date_class = NULL;
 OOP _gst_delay_class = NULL;
 OOP _gst_dictionary_class = NULL;
@@ -122,7 +146,6 @@ OOP _gst_permission_class = NULL;
 OOP _gst_positionable_stream_class = NULL;
 OOP _gst_process_class = NULL;
 OOP _gst_processor_scheduler_class = NULL;
-OOP _gst_random_class = NULL;
 OOP _gst_read_stream_class = NULL;
 OOP _gst_read_write_stream_class = NULL;
 OOP _gst_root_namespace_class = NULL;
@@ -140,12 +163,18 @@ OOP _gst_sym_link_class = NULL;
 OOP _gst_symbol_class = NULL;
 OOP _gst_system_dictionary_class = NULL;
 OOP _gst_time_class = NULL;
-OOP _gst_token_stream_class = NULL;
 OOP _gst_true_class = NULL;
 OOP _gst_undefined_object_class = NULL;
 OOP _gst_unicode_character_class = NULL;
 OOP _gst_unicode_string_class = NULL;
 OOP _gst_variable_binding_class = NULL;
+OOP _gst_weak_array_class = NULL;
+OOP _gst_weak_set_class = NULL;
+OOP _gst_weak_key_dictionary_class = NULL;
+OOP _gst_weak_value_lookup_table_class = NULL;
+OOP _gst_weak_identity_set_class = NULL;
+OOP _gst_weak_key_identity_dictionary_class = NULL;
+OOP _gst_weak_value_identity_dictionary_class = NULL;
 OOP _gst_write_stream_class = NULL;
 OOP _gst_processor_oop = NULL;
 
@@ -169,8 +198,10 @@ static ssize_t identity_dictionary_find_key (OOP identityDictionaryOOP,
 static size_t identity_dictionary_find_key_or_nil (OOP identityDictionaryOOP,
 						   OOP keyOOP);
 
-/* Create a new instance of IdentityDictionary and answer it.  */
-static OOP identity_dictionary_new (int size);
+/* Create a new instance of CLASSOOP (an IdentityDictionary subclass)
+   and answer it.  */
+static OOP identity_dictionary_new (OOP classOOP,
+				    int size);
 
 /* Create a new instance of Namespace with the given SIZE, NAME and
    superspace (SUPERSPACEOOP).  */
@@ -201,7 +232,7 @@ static void init_smalltalk_dictionary (void);
 
 /* This fills MAP so that it associates primitive numbers in the saved
    image to primitive numbers in this VM.  */
-static void prepare_primitive_numbers_map (int *map);
+static void prepare_primitive_numbers_table (void);
 
 /* Add a global named GLOBALNAME and give it the value GLOBALVALUE.
    Return GLOBALVALUE.  */
@@ -261,11 +292,11 @@ static const char *feature_strings[] = {
 
 static const class_definition class_info[] = {
   {&_gst_object_class, &_gst_nil_oop,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "Object", NULL, "Dependencies FinalizableObjects", "VMPrimitives" },
 
   {&_gst_object_memory_class, &_gst_object_class,
-   ISP_FIXED, 34,
+   ISP_FIXED, true, 34,
    "ObjectMemory", "bytesPerOOP bytesPerOTE "
    "edenSize survSpaceSize oldSpaceSize fixedSpaceSize "
    "edenUsedBytes survSpaceUsedBytes oldSpaceUsedBytes "
@@ -279,404 +310,420 @@ static const class_definition class_info[] = {
    "allocFailures allocMatches allocSplits allocProbes", NULL, NULL },
 
   {&_gst_message_class, &_gst_object_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, true, 2,
    "Message", "selector args", NULL, NULL },
 
   {&_gst_directed_message_class, &_gst_message_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "DirectedMessage", "receiver", NULL, NULL },
 
   {&_gst_magnitude_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "Magnitude", NULL, NULL, NULL },
 
   {&_gst_char_class, &_gst_magnitude_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "Character", "codePoint", "Table UpperTable LowerTable", NULL },
 
   {&_gst_unicode_character_class, &_gst_char_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "UnicodeCharacter", NULL, NULL, NULL },
 
   {&_gst_time_class, &_gst_magnitude_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "Time", "seconds",
    "SecondClockAdjustment ClockOnStartup", NULL },
 
   {&_gst_date_class, &_gst_magnitude_class,
-   ISP_FIXED, 4,
+   ISP_FIXED, false, 4,
    "Date", "days day month year",
    "DayNameDict MonthNameDict", NULL },
 
   {&_gst_number_class, &_gst_magnitude_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "Number", NULL, NULL, NULL },
 
   {&_gst_float_class, &_gst_number_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "Float", NULL, NULL, "CSymbols" },
 
   {&_gst_floatd_class, &_gst_float_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "FloatD", NULL, NULL, "CSymbols" },
 
   {&_gst_floate_class, &_gst_float_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "FloatE", NULL, NULL, "CSymbols" },
 
   {&_gst_floatq_class, &_gst_float_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "FloatQ", NULL, NULL, "CSymbols" },
 
   {&_gst_fraction_class, &_gst_number_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, false, 2,
    "Fraction", "numerator denominator", "Zero One", NULL },
 
   {&_gst_integer_class, &_gst_number_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "Integer", NULL, NULL, "CSymbols" },
 
   {&_gst_small_integer_class, &_gst_integer_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "SmallInteger", NULL, NULL, NULL },
 
   {&_gst_large_integer_class, &_gst_integer_class,	/* these four
 							   classes
 							   added by */
-   ISP_UCHAR, 0,	/* pb Sep 10 18:06:49 1998 */
+   ISP_UCHAR, true, 0,	/* pb Sep 10 18:06:49 1998 */
    "LargeInteger", NULL,
    "Zero One ZeroBytes OneBytes LeadingZeros TrailingZeros", NULL },
 
   {&_gst_large_positive_integer_class, &_gst_large_integer_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "LargePositiveInteger", NULL, NULL, NULL },
 
   {&_gst_large_zero_integer_class, &_gst_large_positive_integer_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "LargeZeroInteger", NULL, NULL, NULL },
 
   {&_gst_large_negative_integer_class, &_gst_large_integer_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "LargeNegativeInteger", NULL, NULL, NULL },
 
   {&_gst_lookup_key_class, &_gst_magnitude_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "LookupKey", "key", NULL, NULL },
 
   {&_gst_association_class, &_gst_lookup_key_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "Association", "value", NULL, NULL },
 
   {&_gst_homed_association_class, &_gst_association_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "HomedAssociation", "environment", NULL, NULL },
 
   {&_gst_variable_binding_class, &_gst_homed_association_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "VariableBinding", NULL, NULL, NULL },
 
   {&_gst_link_class, &_gst_object_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "Link", "nextLink", NULL, NULL },
 
   {&_gst_process_class, &_gst_link_class,
-   ISP_FIXED, 7,
+   ISP_FIXED, true, 7,
    "Process",
    "suspendedContext priority myList name unwindPoints interrupts interruptLock",
    NULL, NULL },
 
   {&_gst_callin_process_class, &_gst_process_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "CallinProcess",
    "returnedValue",
    NULL, NULL },
 
   {&_gst_sym_link_class, &_gst_link_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "SymLink", "symbol", NULL, NULL },
 
   {&_gst_collection_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "Collection", NULL, NULL, NULL },
 
   {&_gst_sequenceable_collection_class, &_gst_collection_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "SequenceableCollection", NULL, NULL, NULL },
 
   {&_gst_linked_list_class, &_gst_sequenceable_collection_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, false, 2,
    "LinkedList", "firstLink lastLink", NULL, NULL },
 
   {&_gst_semaphore_class, &_gst_linked_list_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, true, 2,
    "Semaphore", "signals name", NULL, NULL },
 
   {&_gst_arrayed_collection_class, &_gst_sequenceable_collection_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "ArrayedCollection", NULL, NULL, NULL },
 
   {&_gst_array_class, &_gst_arrayed_collection_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, true, 0,
    "Array", NULL, NULL, NULL },
 
+  {&_gst_weak_array_class, &_gst_array_class,
+   ISP_FIXED, false, 2,
+   "WeakArray", "values nilValues", NULL, NULL },
+
   {&_gst_character_array_class, &_gst_arrayed_collection_class,
-   ISP_ULONG, 0,
+   ISP_ULONG, false, 0,
    "CharacterArray", NULL, NULL, NULL },
 
   {&_gst_string_class, &_gst_character_array_class,
-   ISP_CHARACTER, 0,
+   ISP_CHARACTER, true, 0,
    "String", NULL, NULL, NULL },
 
   {&_gst_unicode_string_class, &_gst_character_array_class,
-   ISP_UTF32, 0,
+   ISP_UTF32, true, 0,
    "UnicodeString", NULL, NULL, NULL },
 
   {&_gst_symbol_class, &_gst_string_class,
-   ISP_CHARACTER, 0,
+   ISP_CHARACTER, true, 0,
    "Symbol", NULL, NULL, NULL },
 
   {&_gst_byte_array_class, &_gst_arrayed_collection_class,
-   ISP_UCHAR, 0,
+   ISP_UCHAR, true, 0,
    "ByteArray", NULL, NULL, "CSymbols" },
 
   {&_gst_compiled_code_class, &_gst_arrayed_collection_class,
-   ISP_UCHAR, 2,
+   ISP_UCHAR, false, 2,
    "CompiledCode", "literals header",
    NULL, NULL },
 
   {&_gst_compiled_block_class, &_gst_compiled_code_class,
-   ISP_UCHAR, 1,
+   ISP_UCHAR, true, 1,
    "CompiledBlock", "method",
    NULL, NULL },
 
   {&_gst_compiled_method_class, &_gst_compiled_code_class,
-   ISP_UCHAR, 1,
+   ISP_UCHAR, true, 1,
    "CompiledMethod", "descriptor ",
    NULL, NULL },
 
   {&_gst_interval_class, &_gst_arrayed_collection_class,
-   ISP_FIXED, 3,
+   ISP_FIXED, true, 3,
    "Interval", "start stop step", NULL, NULL },
 
   {&_gst_ordered_collection_class, &_gst_sequenceable_collection_class,
-   ISP_POINTER, 2,
+   ISP_POINTER, false, 2,
    "OrderedCollection", "firstIndex lastIndex", NULL, NULL },
 
   {&_gst_sorted_collection_class, &_gst_ordered_collection_class,
-   ISP_POINTER, 3,
+   ISP_POINTER, false, 3,
    "SortedCollection", "lastOrdered sorted sortBlock",
    "DefaultSortBlock",
    NULL },
 
   {&_gst_bag_class, &_gst_collection_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, false, 1,
    "Bag", "contents", NULL, NULL },
 
   {&_gst_mapped_collection_class, &_gst_collection_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, false, 2,
    "MappedCollection", "domain map", NULL, NULL },
 
   {&_gst_hashed_collection_class, &_gst_collection_class,
-   ISP_POINTER, 1,
+   ISP_POINTER, false, 1,
    "HashedCollection", "tally", NULL, NULL },
 
   {&_gst_set_class, &_gst_hashed_collection_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "Set", NULL, NULL, NULL },
 
+  {&_gst_weak_set_class, &_gst_set_class,
+   ISP_POINTER, false, 0,
+   "WeakSet", NULL, NULL, NULL },
+
   {&_gst_identity_set_class, &_gst_set_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "IdentitySet", NULL, NULL, NULL },
 
+  {&_gst_weak_identity_set_class, &_gst_weak_set_class,
+   ISP_POINTER, false, 0,
+   "WeakIdentitySet", NULL, NULL, NULL },
+
   {&_gst_dictionary_class, &_gst_hashed_collection_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, true, 0,
    "Dictionary", NULL, NULL, NULL },
 
+  {&_gst_weak_key_dictionary_class, &_gst_dictionary_class,
+   ISP_POINTER, false, 1,
+   "WeakKeyDictionary", "keys", NULL, NULL },
+
+  {&_gst_weak_key_identity_dictionary_class, &_gst_weak_key_dictionary_class,
+   ISP_POINTER, false, 0,
+   "WeakKeyIdentityDictionary", NULL, NULL, NULL },
+
   {&_gst_lookup_table_class, &_gst_dictionary_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "LookupTable", NULL, NULL, NULL },
 
+  {&_gst_weak_value_lookup_table_class, &_gst_lookup_table_class,
+   ISP_POINTER, false, 1,
+   "WeakValueLookupTable", "values", NULL, NULL },
+
+  {&_gst_weak_value_identity_dictionary_class, &_gst_weak_value_lookup_table_class,
+   ISP_POINTER, false, 0,
+   "WeakValueIdentityDictionary", NULL, NULL, NULL },
+
   {&_gst_identity_dictionary_class, &_gst_lookup_table_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "IdentityDictionary", NULL, NULL, NULL },
 
   {&_gst_method_dictionary_class, &_gst_lookup_table_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, true, 0,
    "MethodDictionary", NULL, NULL, NULL },
 
   /* These five MUST have the same structure as dictionary; they're
      used interchangeably within the C portion of the system */
   {&_gst_binding_dictionary_class, &_gst_dictionary_class,
-   ISP_POINTER, 1,
+   ISP_POINTER, true, 1,
    "BindingDictionary", "environment", NULL, NULL },
 
   {&_gst_abstract_namespace_class, &_gst_binding_dictionary_class,
-   ISP_POINTER, 2,
+   ISP_POINTER, true, 2,
    "AbstractNamespace", "name subspaces", NULL, NULL },
 
   {&_gst_root_namespace_class, &_gst_abstract_namespace_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "RootNamespace", NULL, NULL, NULL },
 
   {&_gst_namespace_class, &_gst_abstract_namespace_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, true, 0,
    "Namespace", NULL, "Current", NULL },
 
   {&_gst_system_dictionary_class, &_gst_root_namespace_class,
-   ISP_POINTER, 0,
+   ISP_POINTER, false, 0,
    "SystemDictionary", NULL, NULL, NULL },
 
   {&_gst_stream_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "Stream", NULL, NULL, NULL },
 
-  {&_gst_token_stream_class, &_gst_stream_class,
-   ISP_FIXED, 1,
-   "TokenStream", "charStream", NULL, NULL },
-
   {&_gst_positionable_stream_class, &_gst_stream_class,
-   ISP_FIXED, 4,
+   ISP_FIXED, false, 4,
    "PositionableStream", "collection ptr endPtr access", NULL, NULL },
 
   {&_gst_read_stream_class, &_gst_positionable_stream_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "ReadStream", NULL, NULL, NULL },
 
   {&_gst_write_stream_class, &_gst_positionable_stream_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "WriteStream", NULL, NULL, NULL },
 
   {&_gst_read_write_stream_class, &_gst_write_stream_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "ReadWriteStream", NULL, NULL, NULL },
 
   {&_gst_byte_stream_class, &_gst_read_write_stream_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "ByteStream", NULL, NULL, NULL },
 
   {&_gst_file_descriptor_class, &_gst_byte_stream_class,
-   ISP_FIXED, 5,
-   "FileDescriptor", "file name isPipe atEnd peek", NULL, NULL },
+   ISP_FIXED, true, 5,
+   "FileDescriptor", "file name isPipe atEnd peek", "AllOpenFiles", NULL },
 
   {&_gst_file_stream_class, &_gst_file_descriptor_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, true, 2,
    "FileStream", "writePtr writeEnd", "Verbose Record Includes", NULL },
 
-  {&_gst_random_class, &_gst_stream_class,
-   ISP_FIXED, 1,
-   "Random", "seed", "Source", NULL },
-
   {&_gst_undefined_object_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "UndefinedObject", NULL, NULL, NULL },
 
   {&_gst_boolean_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "Boolean", NULL, NULL, NULL },
 
   {&_gst_false_class, &_gst_boolean_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "False", "truthValue", NULL, NULL },
 
   {&_gst_true_class, &_gst_boolean_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "True", "truthValue", NULL, NULL },
 
   {&_gst_processor_scheduler_class, &_gst_object_class,
-   ISP_FIXED, 6,
+   ISP_FIXED, false, 6,
    "ProcessorScheduler",
    "processLists activeProcess idleTasks processTimeslice gcSemaphore gcArray",
    NULL, NULL },
 
   {&_gst_delay_class, &_gst_object_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, false, 2,
    "Delay", "resumptionTime isRelative",
    "Queue TimeoutSem MutexSem DelayProcess IdleProcess", NULL },
 
   {&_gst_shared_queue_class, &_gst_object_class,
-   ISP_FIXED, 3,
+   ISP_FIXED, false, 3,
    "SharedQueue", "queueSem valueReady queue", NULL, NULL },
 
   /* Change this, classDescription, or gst_class, and you must change
      the implementaion of new_metaclass some */
   {&_gst_behavior_class, &_gst_object_class,
-   ISP_FIXED, 5,
+   ISP_FIXED, true, 5,
    "Behavior",
    "superClass subClasses methodDictionary instanceSpec instanceVariables",
    NULL, NULL },
 
   {&_gst_class_description_class, &_gst_behavior_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, true, 0,
    "ClassDescription", NULL, NULL, NULL },
 
   {&_gst_class_class, &_gst_class_description_class,
-   ISP_FIXED, 8,
+   ISP_FIXED, true, 8,
    "Class",
    "name comment category environment classVariables sharedPools "
    "securityPolicy pragmaHandlers",
    NULL, NULL },
 
   {&_gst_metaclass_class, &_gst_class_description_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "Metaclass", "instanceClass", NULL, NULL },
 
   {&_gst_context_part_class, &_gst_object_class,
-   ISP_POINTER, 6,
+   ISP_POINTER, true, 6,
    "ContextPart", "parent nativeIP ip sp receiver method ",
-   "UnwindPoints", NULL },
+   NULL, NULL },
 
   {&_gst_method_context_class, &_gst_context_part_class,
-   ISP_POINTER, 1,
+   ISP_POINTER, true, 1,
    "MethodContext", "flags ", NULL, NULL },
 
   {&_gst_block_context_class, &_gst_context_part_class,
-   ISP_POINTER, 1,
+   ISP_POINTER, true, 1,
    "BlockContext", "outerContext ", NULL, NULL },
 
+  {&_gst_continuation_class, &_gst_object_class,
+   ISP_FIXED, true, 1,
+   "Continuation", "stack ", NULL, NULL },
+
   {&_gst_block_closure_class, &_gst_object_class,
-   ISP_FIXED, 3,
+   ISP_FIXED, true, 3,
    "BlockClosure", "outerContext block receiver", NULL, NULL },
 
-
-/***********************************************************************
- *
- *	End of Standard Smalltalk Class definitions.  The definitions below are
- *	specific to GNU Smalltalk.
- *
- ***********************************************************************/
-
   {&_gst_permission_class, &_gst_object_class,
-   ISP_FIXED, 4,
+   ISP_FIXED, true, 4,
    "Permission", "name actions target positive", NULL, NULL },
 
   {&_gst_security_policy_class, &_gst_object_class,
-   ISP_FIXED, 2,
+   ISP_FIXED, true, 2,
    "SecurityPolicy", "dictionary owner", NULL, NULL },
 
   {&_gst_c_object_class, &_gst_object_class,
-   ISP_ULONG, 1,	/* leave this this way */
+   ISP_ULONG, true, 1,	/* leave this this way */
    "CObject", "type", NULL, "CSymbols" },
 
   {&_gst_c_type_class, &_gst_object_class,
-   ISP_FIXED, 1,
+   ISP_FIXED, true, 1,
    "CType", "cObjectType", NULL, NULL },
 
   {&_gst_c_func_descriptor_class, &_gst_object_class,
-   ISP_POINTER, 4,
+   ISP_POINTER, true, 4,
    "CFunctionDescriptor",
    "cFunction cFunctionName returnType numFixedArgs",
    NULL, NULL },
 
   {&_gst_memory_class, &_gst_object_class,
-   ISP_FIXED, 0,
+   ISP_FIXED, false, 0,
    "Memory", NULL, NULL, NULL },
 
   {&_gst_method_info_class, &_gst_object_class,
-   ISP_POINTER, 4,
+   ISP_POINTER, true, 4,
    "MethodInfo", "sourceCode category class selector", NULL, NULL },
 
   {&_gst_file_segment_class, &_gst_object_class,
-   ISP_FIXED, 3,
+   ISP_FIXED, true, 3,
    "FileSegment", "file startPos size", NULL, NULL }
 
 /* Classes not defined here (like Point/Rectangle/RunArray) are
@@ -727,7 +774,7 @@ init_proto_oops()
 
   smalltalkDictionary->objClass = _gst_system_dictionary_class;
   smalltalkDictionary->tally = FROM_INT(0);
-  smalltalkDictionary->name = _gst_intern_string ("Smalltalk");
+  smalltalkDictionary->name = _gst_smalltalk_namespace_symbol;
   smalltalkDictionary->superspace = _gst_nil_oop;
   smalltalkDictionary->subspaces = _gst_nil_oop;
   nil_fill (smalltalkDictionary->assoc,
@@ -746,23 +793,25 @@ init_proto_oops()
 void
 _gst_init_dictionary (void)
 {
+  memcpy (_gst_primitive_table, _gst_default_primitive_table,
+          sizeof (_gst_primitive_table));
+
   /* The order of this must match the indices defined in oop.h!! */
   _gst_smalltalk_dictionary = alloc_oop (NULL, _gst_mem.active_flag);
   _gst_processor_oop = alloc_oop (NULL, _gst_mem.active_flag);
   _gst_symbol_table = alloc_oop (NULL, _gst_mem.active_flag);
 
+  _gst_init_symbols_pass1 ();
+
   create_classes_pass1 (class_info, sizeof (class_info) / sizeof (class_info[0]));
 
   init_proto_oops();
+  _gst_init_symbols_pass2 ();
   init_smalltalk_dictionary ();
 
   create_classes_pass2 (class_info, sizeof (class_info) / sizeof (class_info[0]));
 
-  add_smalltalk ("KernelFilePath",
-		 _gst_string_new (_gst_kernel_file_path));
-
   init_runtime_objects ();
-
   _gst_tenure_all_survivors ();
 }
 
@@ -967,14 +1016,13 @@ init_smalltalk_dictionary (void)
 
   add_smalltalk ("Smalltalk", _gst_smalltalk_dictionary);
   add_smalltalk ("Version", _gst_string_new (fullVersionString));
+  add_smalltalk ("KernelFilePath", _gst_string_new (_gst_kernel_file_path));
   add_smalltalk ("CObjectType", _gst_c_object_type_ctype);
   add_smalltalk ("KernelInitialized", _gst_false_oop);
   add_smalltalk ("SymbolTable", _gst_symbol_table);
   add_smalltalk ("Processor", _gst_processor_oop);
   add_smalltalk ("Features", featuresArrayOOP);
 
-  _gst_init_symbols ();
-
   /* Add subspaces */
   add_smalltalk ("CSymbols",
     namespace_new (32, "CSymbols", _gst_smalltalk_dictionary));
@@ -1006,16 +1054,13 @@ add_smalltalk (const char *globalName,
 void
 init_runtime_objects (void)
 {
-  add_smalltalk ("KernelFileSystemPath", _gst_string_new (KERNEL_PATH));
+  add_smalltalk ("UserFileBasePath", _gst_string_new (_gst_user_file_base_path));
   add_smalltalk ("ModulePath", _gst_string_new (MODULE_PATH));
-  add_smalltalk ("KernelFileLocalPath",
-		 _gst_string_new (_gst_kernel_file_path));
-  add_smalltalk ("ImageFilePath",
-		 _gst_string_new (_gst_image_file_path));
-  add_smalltalk ("ImageFileName",
-		 _gst_string_new (_gst_binary_image_name));
-  add_smalltalk ("OutputVerbosity",
-		 FROM_INT (_gst_verbosity));
+  add_smalltalk ("LibexecPath", _gst_string_new (LIBEXEC_PATH));
+  add_smalltalk ("ImageFilePath", _gst_string_new (_gst_image_file_path));
+  add_smalltalk ("ExecutableFileName", _gst_string_new (_gst_executable_path));
+  add_smalltalk ("ImageFileName", _gst_string_new (_gst_binary_image_name));
+  add_smalltalk ("OutputVerbosity", FROM_INT (_gst_verbosity));
   add_smalltalk ("RegressionTesting",
 		 _gst_regression_testing ? _gst_true_oop : _gst_false_oop);
 
@@ -1066,6 +1111,14 @@ init_c_symbols ()
 #define NAN (0.0 / 0.0)
 #endif
 
+#if defined WIN32 && !defined __CYGWIN__
+  NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("PathSeparator"),
+		    CHAR_OOP_AT ('\\'));
+#else
+  NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("PathSeparator"),
+		    CHAR_OOP_AT ('/'));
+#endif
+
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleMin"),
 		    floatd_new (DBL_MIN));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleMax"),
@@ -1077,7 +1130,7 @@ init_c_symbols ()
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleNaN"),
 		    floatd_new ((double) NAN));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleDigits"),
-		    FROM_INT (DBL_DIG));
+		    FROM_INT (ceil (DBL_MANT_DIG * 0.301029995663981)));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleBinaryDigits"),
 		    FROM_INT (DBL_MANT_DIG));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CDoubleMinExp"),
@@ -1098,7 +1151,7 @@ init_c_symbols ()
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CFloatMax"),
 		    floate_new (FLT_MAX));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CFloatDigits"),
-		    FROM_INT (FLT_DIG));
+		    FROM_INT (ceil (FLT_MANT_DIG * 0.301029995663981)));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CFloatBinaryDigits"),
 		    FROM_INT (FLT_MANT_DIG));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CFloatMinExp"),
@@ -1119,7 +1172,7 @@ init_c_symbols ()
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CLongDoubleMax"),
 		    floatq_new (LDBL_MAX));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CLongDoubleDigits"),
-		    FROM_INT (LDBL_DIG));
+		    FROM_INT (ceil (LDBL_MANT_DIG * 0.301029995663981)));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CLongDoubleBinaryDigits"),
 		    FROM_INT (LDBL_MANT_DIG));
   NAMESPACE_AT_PUT (cSymbolsOOP, _gst_intern_string ("CLongDoubleMinExp"),
@@ -1203,11 +1256,9 @@ create_class (const class_definition *ci)
 
 
 mst_Boolean
-_gst_init_dictionary_on_image_load (size_t numOOPs)
+_gst_init_dictionary_on_image_load (mst_Boolean prim_table_matches)
 {
   const class_definition *ci;
-  OOP oop;
-  int primitive_numbers_map[NUM_PRIMITIVES];
 
   _gst_smalltalk_dictionary = OOP_AT (SMALLTALK_OOP_INDEX);
   _gst_processor_oop = OOP_AT (PROCESSOR_OOP_INDEX);
@@ -1217,66 +1268,56 @@ _gst_init_dictionary_on_image_load (size_t numOOPs)
       || IS_NIL (_gst_smalltalk_dictionary))
     return (false);
 
+  _gst_restore_symbols ();
+
   for (ci = class_info; 
        ci < class_info + sizeof(class_info) / sizeof(class_definition);
        ci++)
-    {
-      *ci->classVar = dictionary_at (_gst_smalltalk_dictionary,
-				     _gst_intern_string (ci->name));
-      if UNCOMMON (IS_NIL (*ci->classVar))
-	return (false);
-    }
+    if (ci->reloadAddress)
+      {
+	*ci->classVar = dictionary_at (_gst_smalltalk_dictionary,
+				       _gst_intern_string (ci->name));
+        if UNCOMMON (IS_NIL (*ci->classVar))
+	  return (false);
+      }
 
   _gst_current_namespace =
-    dictionary_at (_gst_class_variable_dictionary
-		   (_gst_namespace_class),
+    dictionary_at (_gst_class_variable_dictionary (_gst_namespace_class),
 		   _gst_intern_string ("Current"));
 
   _gst_c_object_type_ctype = dictionary_at (_gst_smalltalk_dictionary,
-					    _gst_intern_string
-					    ("CObjectType"));
+					    _gst_intern_string ("CObjectType"));
 
   _gst_init_builtin_objects_classes ();
-  _gst_init_symbols ();
 
-  /* Important: this is called *before* init_primitives_dictionary
-     fills the VMPrimitives dictionary and *after* _gst_init_symbols
+  /* Important: this is called *after* _gst_init_symbols
      fills in _gst_vm_primitives_symbol! */
-  prepare_primitive_numbers_map (primitive_numbers_map);
+  if (prim_table_matches)
+    memcpy (_gst_primitive_table, _gst_default_primitive_table,
+            sizeof (_gst_primitive_table));
+  else
+    prepare_primitive_numbers_table ();
 
-  init_primitives_dictionary ();
   init_runtime_objects ();
-
-  for (oop = _gst_mem.ot_base; oop < &_gst_mem.ot[numOOPs]; oop++)
-    {
-      if (!IS_OOP_VALID_GC (oop))
-	continue;
-
-      if UNCOMMON (OOP_CLASS (oop) == _gst_c_func_descriptor_class)
-	_gst_restore_cfunc_descriptor (oop); /* in cint.c */
-      else if UNCOMMON (OOP_CLASS (oop) == _gst_compiled_method_class)
-	_gst_restore_primitive_number (oop, primitive_numbers_map);
-      else if UNCOMMON (OOP_CLASS (oop) == _gst_callin_process_class)
-	_gst_terminate_process (oop);
-    }
-
   return (true);
 }
 
 void
-prepare_primitive_numbers_map (int *map)
+prepare_primitive_numbers_table ()
 {
   int i;
   OOP primitivesDictionaryOOP;
   gst_dictionary primitivesDictionary;
 
-  memzero (map, NUM_PRIMITIVES * sizeof (int));
   primitivesDictionaryOOP = dictionary_at (_gst_smalltalk_dictionary, 
 					   _gst_vm_primitives_symbol);
 
   primitivesDictionary =
     (gst_dictionary) OOP_TO_OBJ (primitivesDictionaryOOP);
 
+  for (i = 0; i < NUM_PRIMITIVES; i++)
+    _gst_set_primitive_attributes (i, NULL);
+
   for (i = 0; i < NUM_PRIMITIVES; i++)
     {
       prim_table_entry *pte = _gst_get_primitive_attributes (i);
@@ -1289,15 +1330,14 @@ prepare_primitive_numbers_map (int *map)
       symbolOOP = _gst_intern_string (pte->name);
       valueOOP = dictionary_at (primitivesDictionaryOOP, symbolOOP);
 
-      if IS_NIL (valueOOP)
+      if (IS_NIL (valueOOP))
         {
           _gst_errorf ("bad primitive name");
           continue;
         }
 
       old_index = TO_INT (valueOOP);
-      map[old_index] = i;
-      /* printf ("Old primitive %d is now %d\n", old_index, i); */
+      _gst_set_primitive_attributes (old_index, pte);
     }
 }
 
@@ -1335,9 +1375,8 @@ _gst_valid_class_method_dictionary (OOP class_oop)
     {
       OOP identDict;
       gst_object obj;
-      identDict = identity_dictionary_new (32);
+      identDict = identity_dictionary_new (_gst_method_dictionary_class, 32);
       obj = OOP_TO_OBJ (identDict);
-      obj->objClass = _gst_method_dictionary_class;
       class = (gst_class) OOP_TO_OBJ (class_oop);
       class->methodDictionary = identDict;
     }
@@ -1404,8 +1443,8 @@ _gst_shared_pool_dictionary (OOP class_oop)
 
 
 OOP
-_gst_find_shared_pool_variable (OOP poolOOP,
-				OOP symbol)
+_gst_namespace_association_at (OOP poolOOP,
+			       OOP symbol)
 {
   OOP assocOOP;
   gst_namespace pool;
@@ -1430,6 +1469,17 @@ _gst_find_shared_pool_variable (OOP poolOOP,
       poolOOP = pool->superspace;
     }
 }
+
+OOP
+_gst_namespace_at (OOP poolOOP,
+		   OOP symbol)
+{
+  OOP assocOOP = _gst_namespace_association_at (poolOOP, symbol);
+  if (IS_NIL (assocOOP))
+    return assocOOP;
+  else
+    return ASSOCIATION_VALUE (assocOOP);
+}
 
 
 size_t
@@ -1508,9 +1558,9 @@ _gst_grow_dictionary (OOP oldDictionaryOOP)
      object, the new dictionary itself */
 
   dictionary = instantiate_with (OOP_CLASS (oldDictionaryOOP), 
-						  numFields, &dictionaryOOP);
+				 numFields, &dictionaryOOP);
   memcpy (dictionary->data, oldDictionary->data, sizeof (PTR) * numFixedFields);
-  oldDictionary = OOP_TO_OBJ (dictionaryOOP);
+  oldDictionary = OOP_TO_OBJ (oldDictionaryOOP);
 
   /* rehash all associations from old dictionary into new one */
   for (i = 0; i < oldNumFields; i++)
@@ -1641,7 +1691,7 @@ identity_dictionary_find_key_or_nil (OOP identityDictionaryOOP,
 }
 
 OOP
-identity_dictionary_new (int size)
+identity_dictionary_new (OOP classOOP, int size)
 {
   gst_identity_dictionary identityDictionary;
   OOP identityDictionaryOOP;
@@ -1649,8 +1699,7 @@ identity_dictionary_new (int size)
   size = new_num_fields (size);
 
   identityDictionary = (gst_identity_dictionary)
-    instantiate_with (_gst_identity_dictionary_class, size * 2,
-		      &identityDictionaryOOP);
+    instantiate_with (classOOP, size * 2, &identityDictionaryOOP);
 
   identityDictionary->tally = FROM_INT (0);
   return (identityDictionaryOOP);
@@ -2014,24 +2063,6 @@ _gst_message_new_args (OOP selectorOOP,
   return (messageOOP);
 }
 
-OOP
-_gst_directed_message_new_args (OOP receiverOOP,
-				OOP selectorOOP,
-			        OOP argsArray)
-{
-  gst_directed_message dirMessage;
-  OOP dirMessageOOP;
-
-  dirMessage = (gst_directed_message) new_instance (_gst_directed_message_class,
-						    &dirMessageOOP);
-
-  dirMessage->selector = selectorOOP;
-  dirMessage->args = argsArray;
-  dirMessage->receiver = receiverOOP;
-
-  return (dirMessageOOP);
-}
-
 OOP
 _gst_c_object_new_typed (PTR cObjPtr,
 			 OOP typeOOP)
