@@ -10,7 +10,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -21,6 +21,28 @@
  * Software Foundation; either version 2, or (at your option) any later
  * version.
  *
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -118,7 +140,7 @@ interp_jmp_buf;
 /* If this is true, for each byte code that is executed, we print on
    stdout the byte index within the current gst_compiled_method and a
    decoded interpretation of the byte code.  */
-mst_Boolean _gst_execution_tracing = false;
+int _gst_execution_tracing = 0;
 
 /* When this is true, and an interrupt occurs (such as SIGABRT),
    Smalltalk will terminate itself by making a core dump (normally it
@@ -130,7 +152,8 @@ mst_Boolean _gst_make_core_file = false;
 mst_Boolean _gst_non_interactive = true;
 
 /* The table of functions that implement the primitives.  */
-static prim_table_entry _gst_primitive_table[NUM_PRIMITIVES];
+prim_table_entry _gst_primitive_table[NUM_PRIMITIVES];
+prim_table_entry _gst_default_primitive_table[NUM_PRIMITIVES];
 
 /* Some performance counters from the interpreter: these
    count the number of special returns.  */
@@ -254,7 +277,7 @@ static interp_jmp_buf *reentrancy_jmp_buf = NULL;
 
 /* when this flag is on and execution tracing is in effect, the top of
    the stack is printed as well as the byte code */
-static mst_Boolean verbose_exec_tracing = false;
+static int verbose_exec_tracing = false;
 
 /* This is the bridge to the primitive operations in the GNU Smalltalk
    system.  This function invokes the proper primitive_func with the
@@ -328,10 +351,13 @@ static void sleep_process (OOP processOOP);
    processes for PROCESSOOP's priority.  Return PROCESSOOP.  */
 static OOP activate_process (OOP processOOP);
 
+/* Restore the virtual machine's state from the ContextPart OOP.  */
+static void resume_suspended_context (OOP oop);
+
 /* Save the virtual machine's state into the suspended Process and
    ContextPart objects, and load them from NEWPROCESS and from
    NEWPROCESS's suspendedContext.  The Processor (the only instance
-   of ProcessorScheduler is also updated accordingly.  */
+   of ProcessorScheduler) is also updated accordingly.  */
 static void change_process_context (OOP newProcess);
 
 /* Mark the semaphores attached to the process system (asynchronous
@@ -448,11 +474,6 @@ static inline void prepare_context (gst_context_part context,
    status (ip, sp, _gst_this_method, _gst_self, ...).  */
 static void unwind_context (void);
 
-/* Used to help minimize the number of primitives used to control the
-   various debugging flags, this routine maps the variable's INDEX to the
-   address of a boolean debug flag, which it returns.  */
-static inline mst_Boolean *bool_addr_index (int index) ATTRIBUTE_PURE;
-
 /* Check whether it is true that sending SENDSELECTOR to RECEIVER
    accepts NUMARGS arguments.  Note that the RECEIVER is only used to
    do a quick check in the method cache before examining the selector
@@ -1298,7 +1319,6 @@ _gst_make_block_closure (OOP blockOOP)
 void
 change_process_context (OOP newProcess)
 {
-  gst_method_context thisContext;
   OOP processOOP;
   gst_process process;
   gst_processor_scheduler processor;
@@ -1329,7 +1349,7 @@ change_process_context (OOP newProcess)
       disable_preemption = !IS_NIL (process->interrupts) &&
 		           TO_INT (process->interrupts) < 0;
 
-      _gst_this_context_oop = process->suspendedContext;
+      resume_suspended_context (process->suspendedContext);
 
       /* Interrupt-enabling cannot be controlled globally from Smalltalk,
          but only on a per-Process basis.  You might think that this leaves
@@ -1354,8 +1374,15 @@ change_process_context (OOP newProcess)
             _gst_enable_interrupts();
 	}
     }
+}
 
-  thisContext = (gst_method_context) OOP_TO_OBJ (_gst_this_context_oop);
+void
+resume_suspended_context (OOP oop)
+{
+  gst_method_context thisContext;
+
+  _gst_this_context_oop = oop;
+  thisContext = (gst_method_context) OOP_TO_OBJ (oop);
   SET_THIS_METHOD (thisContext->method, GET_CONTEXT_IP (thisContext));
   sp = thisContext->contextStack + TO_INT (thisContext->spOffset);
 
@@ -1365,6 +1392,7 @@ change_process_context (OOP newProcess)
 
   _gst_temporaries = thisContext->contextStack;
   _gst_self = thisContext->receiver;
+  free_lifo_context = lifo_contexts;
 }
 
 
@@ -1938,8 +1966,6 @@ semaphore_new (int signals)
   return (semaphoreOOP);
 }
 
-/* runs before every evaluation (_gst_execute_statements) and before GC turned on.
-   Note that we don't use the incubator because _gst_processor_oop is a global.  */
 void
 _gst_init_process_system (void)
 {
@@ -2004,25 +2030,66 @@ create_callin_process (OOP contextOOP)
   return (initialProcessOOP);
 }
 
-mst_Boolean *
-bool_addr_index (int index)
+int
+_gst_get_var (enum gst_var_index index)
 {
   switch (index)
     {
-    case 0:
-      return (&_gst_declare_tracing);
-    case 1:
-      return (&_gst_execution_tracing);
-    case 2:
-      return (&verbose_exec_tracing);
-    case 3:
-      return (&_gst_gc_message);
+    case GST_DECLARE_TRACING:
+      return (_gst_declare_tracing);
+    case GST_EXECUTION_TRACING:
+      return (_gst_execution_tracing);
+    case GST_EXECUTION_TRACING_VERBOSE:
+      return (verbose_exec_tracing);
+    case GST_GC_MESSAGE:
+      return (_gst_gc_message);
+    case GST_VERBOSITY:
+      return (_gst_verbosity);
+    case GST_MAKE_CORE_FILE:
+      return (_gst_make_core_file);
+    case GST_REGRESSION_TESTING:
+      return (_gst_regression_testing);
     default:
-      return (NULL);		/* index out of range, signal the error 
-				 */
+      return (-1);
     }
 }
 
+int
+_gst_set_var (enum gst_var_index index, int value)
+{
+  int old = _gst_get_var (index);
+  if (value < 0)
+    return -1;
+
+  switch (index)
+    {
+    case GST_DECLARE_TRACING:
+      _gst_declare_tracing = value;
+      break;
+    case GST_EXECUTION_TRACING:
+      _gst_execution_tracing = value;
+      break;
+    case GST_EXECUTION_TRACING_VERBOSE:
+      verbose_exec_tracing = value;
+      break;
+    case GST_GC_MESSAGE:
+      _gst_gc_message = value;
+      break;
+    case GST_VERBOSITY:
+      _gst_verbosity = value;
+      break;
+    case GST_MAKE_CORE_FILE:
+      _gst_make_core_file = value;
+      break;
+    case GST_REGRESSION_TESTING:
+      _gst_regression_testing = true;
+      break;
+    default:
+      return (-1);
+    }
+
+  return old;
+}
 
 
 void
@@ -2105,25 +2172,6 @@ _gst_nvmsg_send (OOP receiver,
   currentProcessOOP = get_active_process ();
   change_process_context (processOOP);
 
-#if 0
-  if (_gst_kernel_initialized)
-    {
-      gst_array argsArray;
-      OOP argsArrayOOP;
-      argsArray = new_instance_with (_gst_array_class, numArgs, &argsArrayOOP);
-      for (i = 0; i < sendArgs; i++)
-        argsArray->data[i] = args[i];
-
-      INC_ADD_OOP (argsArrayOOP);
-      dirMessageOOP = _gst_directed_message_new_args (receiver, sendSelector, argsArrayOOP);
-
-      receiver = _gst_process_class;
-      sendSelector = _gst_start_execution_symbol;
-      sendArgs = 1;
-      args = &dirMessageOOP;
-    }
-#endif
-
   PUSH_OOP (receiver);
   for (i = 0; i < sendArgs; i++)
     PUSH_OOP (args[i]);
@@ -2248,8 +2296,6 @@ void
 _gst_mark_processor_registers (void)
 {
   mark_semaphore_oops ();
-
-  /* Get everything into the main OOP table first.  */
   if (_gst_this_context_oop)
     MAYBE_MARK_OOP (_gst_this_context_oop);
 
@@ -2442,7 +2488,7 @@ interrupt_handler (int sig)
 #ifdef HAVE_EXECINFO_H
       /* Don't print a backtrace, for example, if exiting during a
 	 compilation.  */
-      if ((_gst_verbosity > 2 && (ip || _gst_gc_running))
+      if ((_gst_verbosity == 3 && (ip || _gst_gc_running))
 	  || is_serious_error || sig == SIGUSR1)
 	{
           PTR array[11];
@@ -2600,7 +2646,16 @@ execute_primitive_operation (int primitive, volatile int numArgs)
 prim_table_entry *
 _gst_get_primitive_attributes (int primitive)
 {
-  return &_gst_primitive_table[primitive];
+  return &_gst_default_primitive_table[primitive];
+}
+
+void
+_gst_set_primitive_attributes (int primitive, prim_table_entry *pte)
+{
+  if (pte)
+    _gst_primitive_table[primitive] = *pte;
+  else
+    _gst_primitive_table[primitive] = _gst_default_primitive_table[0];
 }
 
 void
