@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -59,7 +81,8 @@ enum
   PRIM_IS_PIPE,                 /* no args */
   PRIM_MK_TEMP,                 /* base: */
   PRIM_GET_CHARS_AT,            /* data:from:to:absOfs: */
-  PRIM_PUT_CHARS_AT             /* data:from:to:absOfs: */
+  PRIM_PUT_CHARS_AT,            /* data:from:to:absOfs: */
+  PRIM_SHUTDOWN_WRITE           /* shutdown */
 };
 
 /* These macros are used to quickly compute the number of words needed
@@ -158,6 +181,13 @@ typedef struct gst_block_context
 }
  *gst_block_context;
 
+typedef struct gst_continuation
+{
+  OBJ_HEADER;
+  OOP stack;
+}
+ *gst_continuation;
+
 /* The structure of various objects related to the process system.  */
 typedef struct gst_semaphore
 {
@@ -231,8 +261,9 @@ extern unsigned long _gst_sample_counter
 
 /* If this is true, for each byte code that is executed, we print on
    stdout the byte index within the current gst_compiled_method and a
-   decoded interpretation of the byte code.  */
-extern mst_Boolean _gst_execution_tracing 
+   decoded interpretation of the byte code.  If > 1, it applies also
+   to code not invoked by the user.  */
+extern int _gst_execution_tracing 
   ATTRIBUTE_HIDDEN;
 
 /* When this is true, and an interrupt occurs (such as SIGSEGV),
@@ -404,9 +435,9 @@ extern void _gst_fixup_object_pointers (void)
 extern void _gst_restore_object_pointers (void) 
   ATTRIBUTE_HIDDEN;
 
-/* This runs before every evaluation (_gst_execute_statements) and
-   before GC turned on.  It creates an initial process if no process
-   is ready to run or if no process has been created yet.  */
+/* This runs before every evaluation and before GC turned on.  It creates an
+   initial process if no process is ready to run or if no process has been
+   created yet.  */
 extern void _gst_init_process_system (void) 
   ATTRIBUTE_HIDDEN;
 
@@ -520,14 +551,43 @@ prim_table_entry;
 #define PRIM_RETURN_SMALL_INTEGER	0x0100	/* 31 or 63 bits */
 #define PRIM_RETURN_SMALL_SMALLINTEGER	0x0300	/* 30 or 62 bits */
 
+/* The checksum of the table of primitive numbers.  Right now it is an MD5,
+   computed from part of the C source code of prims.inl.  We compare it when
+   loading an image, to avoid having to reload the primitive table.  */
+extern int _gst_primitives_md5[4];
+
+/* The table of functions that implement the primitives.  */
+extern prim_table_entry _gst_primitive_table[NUM_PRIMITIVES];
+extern prim_table_entry _gst_default_primitive_table[NUM_PRIMITIVES];
+ 
 /* This can be used to obtain information on a particular primitive
    operations in the GNU Smalltalk system.  */
 extern prim_table_entry * _gst_get_primitive_attributes (int primitive)
   ATTRIBUTE_PURE 
   ATTRIBUTE_HIDDEN;
 
+/* Dually, this maps the primitive number that will be used for running
+   the image, to the entry which was returned by _gst_get_primitive_attributes.
+   If PTE is NULL, the primitive will be invalid.  */
+extern void _gst_set_primitive_attributes (int primitive,
+					   prim_table_entry *pte)
+  ATTRIBUTE_HIDDEN;
+
 /* Initialize the table of primitives.  */
 extern void _gst_init_primitives () 
   ATTRIBUTE_HIDDEN;
 
+/* Get the value of internal variable whose number is INDEX; the
+   list of valid variables is in gstpub.h.  Return -1 if the index
+   is invalid.  */
+extern int _gst_get_var (enum gst_var_index index)
+  ATTRIBUTE_HIDDEN;
+
+/* Set the value of internal variable whose number is INDEX; the
+   list of valid variables is in gstpub.h.  Return -1 if the index
+   is invalid or the value is negative, otherwise return the previous
+   value.  */
+extern int _gst_set_var (enum gst_var_index index, int value)
+  ATTRIBUTE_HIDDEN;
+
 #endif /* GST_INTERP_H */
