@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2004,2005,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2004,2005,2006,2007
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -68,6 +90,12 @@ char *_gst_first_error_str = NULL;
 char *_gst_first_error_file = NULL;
 int _gst_first_error_line = 0;
 
+/* Last returned token.  */
+static int last_token;
+
+/* Balance of parentheses.  Used to turn a newline into a period.  */
+static int parenthesis_depth;
+
 /* Answer true if IC is a valid base-10 digit.  */
 static mst_Boolean is_digit (int ic);
 
@@ -76,23 +104,21 @@ static mst_Boolean is_base_digit (int c,
 				  int base);
 
 /* Parse the fractional part of a Float constant.  Store it in
-   NUMPTR.  Read numbers in base-BASE, the first one being C. 
-   Answer the scale (number of digits in numPtr).  If LARGEINTEGER
-   is not NULL, the digits are stored in the buffer maintained
-   by str.c, and LARGEINTEGER is set to true if numPtr does
-   not have sufficient precision.  */
+   NUMPTR.  Read numbers in base-BASE, the first one being C.  Answer the
+   scale (number of digits in numPtr).  If LARGEINTEGER is not NULL,
+   the digits are stored in an obstack, and LARGEINTEGER is set to true
+   if numPtr does not have sufficient precision.  */
 static int scan_fraction (int c,
 			   int base,
 			   long double *numPtr,
 			   mst_Boolean *largeInteger);
 
 /* Parse a numeric constant and return it.  Read numbers in
-   base-BASE, the first one being C.  If a - was parsed,
-   NEGATIVE must be true so that the sign of the result is
-   changed accordingly.  If LARGEINTEGER is not NULL, the
-   digits are stored in the buffer maintained by str.c,
-   and LARGEINTEGER is set to true if the return value does
-   not have sufficient precision.  */
+   base-BASE, the first one being C.  If a - was parsed, NEGATIVE
+   must be true so that the sign of the result is changed accordingly.
+   If LARGEINTEGER is not NULL, the digits are stored in an obstack,
+   and LARGEINTEGER is set to true if the return value does not have
+   sufficient precision.  */
 static long double scan_digits (int c,
 			    mst_Boolean negative,
 			    int base,
@@ -118,9 +144,36 @@ static int comment (int c,
 static int char_literal (int c,
 			 YYSTYPE * lvalp);
 
+/* Remember the current balance of open/close parentheses, used to treat
+   newlines as periods.  */
+static int scan_open_paren (int c,
+			    YYSTYPE * lvalp);
+
+/* Remember the current balance of open/close parentheses, used to treat
+   newlines as periods.  */
+static int scan_close_paren (int c,
+			     YYSTYPE * lvalp);
+
+/* Remember the current balance of open/close parentheses, used to treat
+   newlines as periods.  */
+static int scan_reset_paren (int c,
+			     YYSTYPE * lvalp);
+
+/* If the current balance of open/close parentheses is zero, and the
+   last token was not a period or bang, treat the newline as a period.  */
+static int scan_newline (int c,
+			 YYSTYPE * lvalp);
+
 /* Parse a binary operator.  C is the first symbol in the selector */
 static int scan_bin_op (int c,
-			 YYSTYPE * lvalp);
+			YYSTYPE * lvalp);
+
+/* Actual work for scan_bin_op is done here.  MAYBE_NUMBER is false if
+   we cannot parse a negative number in this context.  */ 
+static int scan_bin_op_1 (int c,
+			  YYSTYPE * lvalp,
+			  mst_Boolean maybe_number);
+
 
 /* Parse a string literal.  C is '\'' */
 static int string_literal (int c,
@@ -147,9 +200,10 @@ static int scan_symbol (int c,
 static int digit_to_int (int c,
 			 int base);
 
-/* Raise BASE to the N-th power.  */
-static inline long double ipowl (long double base,
-			         int n);
+/* Raise BASE to the N-th power and multiply MANT by the result.  */
+static inline long double mul_powl (long double mant,
+				    long double base,
+			            int n);
 
 #ifdef LEXDEBUG
 static void print_token (int token,
@@ -179,7 +233,7 @@ static const lex_tab_elt char_table[128] = {
 /*   7 */ {invalid, 0, 0},
 /*   8 */ {invalid, 0, 0},
 /*   9 */ {0, 0, WHITE_SPACE},
-/*  10 */ {0, 0, WHITE_SPACE},
+/*  10 */ {scan_newline, 0, 0},
 /*  11 */ {invalid, 0, 0},
 /*  12 */ {0, 0, WHITE_SPACE},
 /*  13 */ {0, 0, WHITE_SPACE},
@@ -202,15 +256,15 @@ static const lex_tab_elt char_table[128] = {
 /*  30 */ {invalid, 0, 0},
 /*  31 */ {invalid, 0, 0},
 /*     */ {0, 0, WHITE_SPACE},
-/*   ! */ {0, '!', 0},
+/*   ! */ {scan_reset_paren, 0, 0},
 /*   " */ {comment, 0, 0},
 /*   # */ {scan_symbol, 0, 0},
 /*   $ */ {char_literal, 0, ID_CHAR | SYMBOL_CHAR},
 /*   % */ {scan_bin_op, 0, BIN_OP_CHAR},
 /*   & */ {scan_bin_op, 0, BIN_OP_CHAR},
 /*   ' */ {string_literal, 0, 0},
-/*   ( */ {0, '(', 0},
-/*   ) */ {0, ')', 0},
+/*   ( */ {scan_open_paren, 0, 0},
+/*   ) */ {scan_close_paren, 0, 0},
 /*   * */ {scan_bin_op, 0, BIN_OP_CHAR},
 /*   + */ {scan_bin_op, 0, BIN_OP_CHAR},
 /*   , */ {scan_bin_op, 0, BIN_OP_CHAR},
@@ -260,9 +314,9 @@ static const lex_tab_elt char_table[128] = {
 /*   X */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
 /*   Y */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
 /*   Z */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
-/*   [ */ {0, '[', 0},
+/*   [ */ {scan_open_paren, 0, 0},
 /*   \ */ {scan_bin_op, 0, BIN_OP_CHAR},
-/*   ] */ {0, ']', 0},
+/*   ] */ {scan_close_paren, 0, 0},
 /*   ^ */ {0, '^', 0},
 /*   _ */ {0, ASSIGNMENT, ID_CHAR | SYMBOL_CHAR},
 /*   ` */ {invalid, 0, 0},
@@ -292,9 +346,9 @@ static const lex_tab_elt char_table[128] = {
 /*   x */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
 /*   y */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
 /*   z */ {scan_ident, 0, ID_CHAR | SYMBOL_CHAR},
-/*   { */ {0, '{', 0},
+/*   { */ {scan_open_paren, 0, 0},
 /*   | */ {scan_bin_op, 0, BIN_OP_CHAR},
-/*   } */ {0, '}', 0},
+/*   } */ {scan_close_paren, 0, 0},
 /*   ~ */ {scan_bin_op, 0, BIN_OP_CHAR},
 /*  ^? */ {invalid, 0, 0}
 };
@@ -327,17 +381,19 @@ _gst_yylex (PTR lvalp, YYLTYPE *llocp)
       ct = CHAR_TAB (ic);
       if ((ct->char_class & WHITE_SPACE) == 0)
 	{
-	  _gst_get_location (&llocp->first_column, &llocp->first_line);
-
+	  *llocp = _gst_get_location ();
 	  assert (ct->lexFunc || ct->retToken);
 	  if (ct->lexFunc)
+	    result = (*ct->lexFunc) (ic, (YYSTYPE *) lvalp);
+	  else
+	    result = ct->retToken;
+
+	  if (result)
 	    {
-	      result = (*ct->lexFunc) (ic, (YYSTYPE *) lvalp);
-	      if (result)
-		return (result);
+	      if (_gst_get_cur_stream_prompt ())
+		last_token = result;
+	      return (result);
 	    }
-	  else if (ct->retToken)
-	    return (ct->retToken);
 	}
     }
 
@@ -372,7 +428,50 @@ invalid (int c,
   return (0);			/* tell the lexer to ignore this */
 }
 
+
+int
+scan_reset_paren (int c,
+	         YYSTYPE * lvalp)
+{
+  if (_gst_get_cur_stream_prompt ())
+    parenthesis_depth = 0;
+  return c;
+}
 
+int
+scan_open_paren (int c,
+	         YYSTYPE * lvalp)
+{
+  if (_gst_get_cur_stream_prompt ())
+    parenthesis_depth++;
+  return c;
+}
+
+int
+scan_close_paren (int c,
+	          YYSTYPE * lvalp)
+{
+  if (_gst_get_cur_stream_prompt ())
+    parenthesis_depth--;
+  return c;
+}
+
+int
+scan_newline (int c,
+	      YYSTYPE * lvalp)
+{
+  if (_gst_get_cur_stream_prompt ()
+      && parenthesis_depth == 0
+      && last_token != 0
+      && last_token != '.' && last_token != '!' && last_token != KEYWORD
+      && last_token != BINOP && last_token != '|' && last_token != '<'
+      && last_token != '>' && last_token != ';')
+    return ('.');
+  else
+    return 0;
+}
+
+
 int
 comment (int c,
 	 YYSTYPE * lvalp)
@@ -410,6 +509,12 @@ char_literal (int c,
     }
   else
     {
+      if (ic > 127)
+        {
+          _gst_errorf
+	    ("Invalid character literal, only character codes from 0 to 127 are valid");
+          _gst_had_error = true;
+        }
       lvalp->ival = ic;
       return (CHAR_LITERAL);
     }
@@ -446,9 +551,8 @@ scan_symbol (int c,
   /* Look for a shebang (#! /).  */
   if (ic == '!')
     {
-      int line, column;
-      _gst_get_location (&column, &line);
-      if (line == 1 && column == 2)
+      YYLTYPE loc = _gst_get_location ();
+      if (loc.first_line == 1 && loc.first_column == 2)
         {
           while (((ic = _gst_next_char ()) != EOF)
                  && ic != '\r' && ic != '\n')
@@ -460,7 +564,7 @@ scan_symbol (int c,
   /* We can read a binary operator and return a SYMBOL_LITERAL,... */
   if (CHAR_TAB (ic)->char_class & BIN_OP_CHAR)
     {
-      scan_bin_op (ic, lvalp);
+      scan_bin_op_1 (ic, lvalp, false);
       return SYMBOL_LITERAL;
     }
 
@@ -478,22 +582,23 @@ scan_symbol (int c,
       return '#';
     }
 
-  _gst_reset_buffer ();
-  _gst_add_str_buf_char (ic);
+  obstack_1grow (_gst_compilation_obstack, ic);
 
   while (((ic = _gst_next_char ()) != EOF)
          && (CHAR_TAB (ic)->char_class & SYMBOL_CHAR))
-    _gst_add_str_buf_char (ic);
+    obstack_1grow (_gst_compilation_obstack, ic);
 
   _gst_unread_char (ic);
-  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
+  obstack_1grow (_gst_compilation_obstack, '\0');
+  lvalp->sval = obstack_finish (_gst_compilation_obstack);
   return SYMBOL_LITERAL;
 }
 
 
 int
-scan_bin_op (int c,
-	      YYSTYPE *lvalp)
+scan_bin_op_1 (int c,
+	       YYSTYPE *lvalp,
+	       mst_Boolean maybe_number)
 {
   char buf[3];
   int ic;
@@ -524,7 +629,7 @@ scan_bin_op (int c,
 
       /* We come here also for a negative number, which we handle
          specially.  */
-      if (c == '-' && is_digit (ic))
+      if (maybe_number && c == '-' && is_digit (ic))
 	return (scan_number ('-', lvalp));
 
       buf[1] = 0;
@@ -540,14 +645,19 @@ scan_bin_op (int c,
     return (BINOP);
 }
 
+int
+scan_bin_op (int c,
+	     YYSTYPE *lvalp)
+{
+  return scan_bin_op_1 (c, lvalp, true);
+}
+
 int
 string_literal (int c,
 		YYSTYPE * lvalp)
 {
   int ic;
 
-  _gst_reset_buffer ();
-
   for (;;)
     {
       ic = _gst_next_char ();
@@ -567,9 +677,10 @@ string_literal (int c,
 	      break;
 	    }
 	}
-      _gst_add_str_buf_char (ic);
+      obstack_1grow (_gst_compilation_obstack, ic);
     }
-  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
+  obstack_1grow (_gst_compilation_obstack, '\0');
+  lvalp->sval = obstack_finish (_gst_compilation_obstack);
   return (STRING_LITERAL);
 }
 
@@ -579,14 +690,13 @@ scan_ident (int c,
 {
   int ic, identType;
 
-  _gst_reset_buffer ();
-  _gst_add_str_buf_char (c);
+  obstack_1grow (_gst_compilation_obstack, c);
 
   identType = IDENTIFIER;
 
   while (((ic = _gst_next_char ()) != EOF)
 	 && (CHAR_TAB (ic)->char_class & ID_CHAR))
-    _gst_add_str_buf_char (ic);
+    obstack_1grow (_gst_compilation_obstack, ic);
 
   /* Read a dot as '::' if followed by a letter.  */
   if (ic == '.')
@@ -610,7 +720,7 @@ scan_ident (int c,
 	_gst_unread_char (':');
       else
 	{
-          _gst_add_str_buf_char (':');
+          obstack_1grow (_gst_compilation_obstack, ':');
           identType = KEYWORD;
 	}
     }
@@ -618,28 +728,40 @@ scan_ident (int c,
   else
     _gst_unread_char (ic);
 
-  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
+  obstack_1grow (_gst_compilation_obstack, '\0');
+  lvalp->sval = obstack_finish (_gst_compilation_obstack);
   return (identType);
 }
 
 
 long double
-ipowl (long double base, 
-       int n)
+mul_powl (long double mant, 
+	  long double base,
+	  int n)
 {
-  int k = 1;
   long double result = 1.0;
-  while (n)
+  int n_abs = n < 0 ? -n : n;
+  int exp;
+  int k = 1;
+
+  /* Restrict base to the range from 1.0 to 2.0.  */
+  base = frexpl (base, &exp);
+  base *= 2;
+  exp--;
+
+  while (n_abs)
     {
-      if (n & k)
+      if (n_abs & k)
         {
           result *= base;
-          n ^= k;
+          n_abs ^= k;
         }
       base *= base;
       k <<= 1;
     }
-  return result;
+
+  mant = (n < 0) ? mant / result : mant * result;
+  return ldexpl (mant, exp * n);
 }
 
 int
@@ -663,6 +785,9 @@ scan_number (int c,
       ic = _gst_next_char ();
       if (ic == 'r')
 	{
+          char *p = obstack_finish (_gst_compilation_obstack);
+          obstack_free (_gst_compilation_obstack, p);
+
 	  base = (int) num;
 	  if (base > 36 || largeInteger)
 	    {
@@ -761,8 +886,9 @@ scan_number (int c,
 		(num * (10 raisedToInteger: exponent)
 		   asScaledDecimal: floatExponent) */
 	lvalp->oval =
-	  _gst_msg_send (numOOP, _gst_as_scaled_decimal_scale_symbol,
+	  _gst_msg_send (numOOP, _gst_as_scaled_decimal_radix_scale_symbol,
 			 FROM_INT (exponent),
+			 FROM_INT (base),
 			 FROM_INT ((int) floatExponent), 
 			 NULL);
 
@@ -808,10 +934,8 @@ scan_number (int c,
   else
     _gst_unread_char (ic);
 
-  if (exponent > 0)
-    num *= ipowl ((long double) base, exponent);
-  else
-    num /= ipowl ((long double) base, -exponent);
+  if (exponent != 0)
+    num = mul_powl (num, (long double) base, exponent);
 
   if (isNegative)
     num = -num;
@@ -822,6 +946,8 @@ scan_number (int c,
 
   if (float_type)
     {
+      char *p = obstack_finish (_gst_compilation_obstack);
+      obstack_free (_gst_compilation_obstack, p);
       lvalp->fval = num;
       return (float_type);
     }
@@ -830,7 +956,10 @@ scan_number (int c,
       lvalp->boval = scan_large_integer (isNegative, base);
       return (LARGE_INTEGER_LITERAL);
     }
+  else
     {
+      char *p = obstack_finish (_gst_compilation_obstack);
+      obstack_free (_gst_compilation_obstack, p);
       lvalp->ival = (intptr_t) num;
       return (INTEGER_LITERAL);
     }
@@ -845,9 +974,6 @@ scan_digits (int c,
   long double result;
   mst_Boolean oneDigit = false;
 
-  if (largeInteger)
-    _gst_reset_buffer ();
-
   while (c == '_')
     c = _gst_next_char ();
 
@@ -858,7 +984,7 @@ scan_digits (int c,
       result += digit_to_int (c, base);
       if (largeInteger)
 	{
-	  _gst_add_str_buf_char (digit_to_int (c, base));
+	  obstack_1grow (_gst_compilation_obstack, digit_to_int (c, base));
 	  if (result > -MIN_ST_INT
 	      || (!negative && result > MAX_ST_INT))
 	    *largeInteger = true;
@@ -902,7 +1028,7 @@ scan_fraction (int c,
 
       if (largeInteger)
         {
-          _gst_add_str_buf_char (digit_to_int (c, base));
+          obstack_1grow (_gst_compilation_obstack, digit_to_int (c, base));
           if (num > MAX_ST_INT)
             *largeInteger = true;
         }
@@ -978,13 +1104,14 @@ scan_large_integer (mst_Boolean negative,
   gst_uchar *digits, *result;
   byte_object bo;
 
-  size = _gst_buffer_size ();
+  /* Copy the contents of the currently grown obstack on the stack.  */
+  size = obstack_object_size (_gst_compilation_obstack);
   digits = (gst_uchar *) alloca (size);
-  _gst_copy_buffer (digits);
+  memcpy (digits, obstack_base (_gst_compilation_obstack), size);
 
-  bo =
-    (byte_object) obstack_alloc (_gst_compilation_obstack,
-				 sizeof (struct byte_object) + size);
+  /* And reuse the area on the obstack for a struct byte_object.  */
+  obstack_blank (_gst_compilation_obstack, sizeof (struct byte_object));
+  bo = (byte_object) obstack_finish (_gst_compilation_obstack);
 
   bo->class =
     negative ? _gst_large_negative_integer_class :
