@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -43,6 +65,9 @@
 /* The number of OOPs that are swept on each incremental GC step.  */
 #define INCREMENTAL_SWEEP_STEP	  100
 
+/* Define this flag to turn on debugging dumps for garbage collection */
+/* #define GC_DEBUG_OUTPUT */
+
 /* Define this flag to turn on debugging code for OOP table management */
 /* #define GC_DEBUGGING */
 
@@ -55,14 +80,18 @@
 /* #define NO_SIGSEGV_HANDLING */
 
 /* Define this flag to turn on debugging code for oldspace management */
-/* #define MMAN_DEBUGGING */
+/* #define MMAN_DEBUG_OUTPUT */
 
 #if !defined HAVE_SIGSEGV_RECOVERY || !(HAVE_SIGSEGV_RECOVERY-0)
 #define NO_SIGSEGV_HANDLING
 #endif
 
-#if defined(GC_DEBUGGING)
-#undef OPTIMIZE
+#if defined(GC_DEBUG_OUTPUT)
+#define GC_DEBUGGING
+#endif
+
+#if !defined(OPTIMIZE)
+#define GC_DEBUGGING
 #endif
 
 
@@ -74,7 +103,7 @@ OOP _gst_true_oop = NULL;
 OOP _gst_false_oop = NULL;
 
 /* This is true to show a message whenever a GC happens.  */
-mst_Boolean _gst_gc_message = true;
+int _gst_gc_message = true;
 
 /* This is != 0 in the middle of a GC.  */
 int _gst_gc_running = 0;
@@ -397,13 +426,13 @@ void _gst_update_object_memory_oop (OOP oop)
 }
 
 void
-_gst_init_oop_table (size_t size)
+_gst_init_oop_table (PTR address, size_t size)
 {
   int i;
 
   oop_heap = NULL;
   for (i = MAX_OOP_TABLE_SIZE; i && !oop_heap; i >>= 1)
-    oop_heap = _gst_heap_create (i * sizeof (struct oop_s));
+    oop_heap = _gst_heap_create (address, i * sizeof (struct oop_s));
 
   if (!oop_heap)
     nomemory (true);
@@ -443,7 +472,6 @@ void
 alloc_oop_table (size_t size)
 {
   size_t bytes;
-  OOP oop;
 
   _gst_mem.ot_size = size;
   bytes = (size - FIRST_OOP_INDEX) * sizeof (struct oop_s);
@@ -458,16 +486,6 @@ alloc_oop_table (size_t size)
   _gst_false_oop = &_gst_mem.ot[FALSE_OOP_INDEX];
 
   _gst_mem.num_free_oops = size;
-
-  /* mark the OOPs as available */
-  PREFETCH_START (_gst_mem.ot, PREF_WRITE | PREF_NTA);
-  for (oop = _gst_mem.ot;
-       oop < &_gst_mem.ot[_gst_mem.ot_size]; oop++)
-    {
-      PREFETCH_LOOP (oop, PREF_WRITE | PREF_NTA);
-      oop->flags = F_FREE;
-    }
-
   _gst_mem.first_allocated_oop = _gst_mem.ot;
   _gst_mem.last_allocated_oop = _gst_mem.last_swept_oop = _gst_mem.ot - 1;
   _gst_mem.highest_swept_oop = _gst_mem.ot;
@@ -477,7 +495,6 @@ mst_Boolean
 _gst_realloc_oop_table (size_t newSize)
 {
   size_t bytes;
-  OOP oop;
 
   bytes = (newSize - _gst_mem.ot_size) * sizeof (struct oop_s);
   if (bytes < 0)
@@ -491,15 +508,6 @@ _gst_realloc_oop_table (size_t newSize)
       return (false);
     }
 
-  /* mark the new OOPs as available */
-  PREFETCH_START (&_gst_mem.ot[_gst_mem.ot_size], PREF_WRITE | PREF_NTA);
-  for (oop = &_gst_mem.ot[_gst_mem.ot_size];
-       oop < &_gst_mem.ot[newSize]; oop++)
-    {
-      PREFETCH_LOOP (oop, PREF_WRITE | PREF_NTA);
-      oop->flags = F_FREE;
-    }
-
   _gst_mem.num_free_oops += newSize - _gst_mem.ot_size;
   _gst_mem.ot_size = newSize;
   return (true);
@@ -519,6 +527,55 @@ _gst_dump_oop_table()
           _gst_display_oop_short (oop);
       }
 }
+
+void
+_gst_dump_owners (OOP oop)
+{
+  OOP oop2, lastOOP;
+
+  for (oop2 = _gst_mem.ot_base, lastOOP = &_gst_mem.ot[_gst_mem.ot_size];
+       oop2 < lastOOP; oop2++)
+    if UNCOMMON (IS_OOP_VALID (oop2) && is_owner(oop2, oop))
+      _gst_display_oop (oop2);
+}
+
+void
+_gst_check_oop_table ()
+{
+  OOP oop, lastOOP;
+
+  for (oop = _gst_mem.ot_base, lastOOP = &_gst_mem.ot[_gst_mem.ot_size];
+       oop < lastOOP; oop++)
+    {
+      gst_object object;
+      OOP *scanPtr;
+      int n;
+
+      if (!IS_OOP_VALID_GC (oop))
+	continue;
+
+      object = OOP_TO_OBJ (oop);
+      scanPtr = &object->objClass;
+      if (oop->flags & F_CONTEXT)
+        {
+          gst_method_context ctx;
+          intptr_t methodSP;
+          ctx = (gst_method_context) object;
+          methodSP = TO_INT (ctx->spOffset);
+          n = ctx->contextStack + methodSP + 1 - object->data;
+        }
+      else
+        n = NUM_OOPS (object) + 1;
+
+      while (n--)
+        {
+	  OOP pointedOOP = *scanPtr++;
+	  if (IS_OOP (pointedOOP)
+	      && (!IS_OOP_ADDR (pointedOOP) || !IS_OOP_VALID_GC (pointedOOP)))
+            abort ();
+	}
+    }
+}
 
 
 void
@@ -669,18 +726,22 @@ _gst_make_oop_fixed (OOP oop)
   if (oop->flags & F_FIXED)
     return;
 
-  size = SIZE_TO_BYTES (TO_INT(oop->object->objSize));
-  newObj = (gst_object) _gst_mem_alloc (_gst_mem.fixed, size);
-  if (!newObj)
-    abort ();
+  if ((oop->flags & F_LOADED) == 0)
+    {
+      size = SIZE_TO_BYTES (TO_INT(oop->object->objSize));
+      newObj = (gst_object) _gst_mem_alloc (_gst_mem.fixed, size);
+      if (!newObj)
+        abort ();
 
-  memcpy (newObj, oop->object, size);
-  if (oop->flags & F_OLD)
-    _gst_mem_free (_gst_mem.old, oop->object);
-  else
-    _gst_mem.numOldOOPs++;
+      memcpy (newObj, oop->object, size);
+      if ((oop->flags & F_OLD) == 0)
+	_gst_mem.numOldOOPs++;
+      else
+        _gst_mem_free (_gst_mem.old, oop->object);
+
+      oop->object = newObj;
+    }
 
-  oop->object = newObj;
   oop->flags &= ~(F_SPACES | F_POOLED);
   oop->flags |= F_OLD | F_FIXED;
 }
@@ -700,7 +761,6 @@ _gst_tenure_oop (OOP oop)
         abort ();
 
       memcpy (newObj, oop->object, size);
-
       _gst_mem.numOldOOPs++;
 
       oop->object = newObj;
@@ -811,7 +871,7 @@ reset_survivor_space (surv_space *space)
 void
 oldspace_after_allocating (heap_data *h, heap_block *blk, size_t sz)
 {
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
   printf ("Allocating oldspace page at %p (%d)\n", blk, sz);
 #endif
 
@@ -824,7 +884,7 @@ oldspace_before_freeing (heap_data *h, heap_block *blk, size_t sz)
 {
   grey_area_node *node, *last, **next;
 
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
   printf ("Freeing oldspace page at %p (%d)\n", blk, sz);
 #endif
   
@@ -833,7 +893,7 @@ oldspace_before_freeing (heap_data *h, heap_block *blk, size_t sz)
     if (node->base >= (OOP *)blk
 	&& node->base + node->n <= (OOP *)( ((char *)blk) + sz))
       {
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
 	printf ("  Remembered table entry removed %p..%p\n",
 		node->base, node->base+node->n);
 #endif
@@ -880,7 +940,7 @@ int oldspace_sigsegv_handler (void* fault_address, int serious)
   if (_gst_mem_protect (page, getpagesize(), PROT_READ | PROT_WRITE) == -1 &&
       (errno == ENOMEM || errno == EFAULT || errno == EACCES || errno == EINVAL))
     {
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
       printf ("Plain old segmentation violation -- address = %p\n", page);
 #endif
       reentering = 0;
@@ -891,7 +951,7 @@ int oldspace_sigsegv_handler (void* fault_address, int serious)
   (void) *(volatile char *) fault_address;
   reentering = 0;
 
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
   printf ("Unprotected %p (SIGSEGV at %p)\n", page, fault_address);
 #endif
 
@@ -938,7 +998,7 @@ void
 _gst_compact (size_t new_heap_limit)
 {
   OOP oop;
-  grey_area_node *node, *next;
+  grey_area_node *node, **next, *last;
   heap_data *new_heap = init_old_space (
     new_heap_limit ? new_heap_limit : _gst_mem.old->heap_limit);
 
@@ -956,15 +1016,26 @@ _gst_compact (size_t new_heap_limit)
       update_stats (&stats.timeOfLastCompaction, NULL, NULL);
     }
 
-  /* Discard all the grey pages.  */
-  for (node = _gst_mem.grey_pages.head; node; node = next)
-    {
-      next = node->next;
-      xfree (node);
-    }
-  _gst_mem.grey_pages.head = _gst_mem.grey_pages.tail = NULL;
-  _gst_mem.rememberedTableEntries = 0;
+  /* Leave only pages from the loaded image in the grey table.  */
+  for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
+    if (node->base >= (OOP *)_gst_mem.loaded_base
+        && node->base < _gst_mem.loaded_end)
+      {
+#ifdef MMAN_DEBUG_OUTPUT
+        printf ("  Remembered table entry left for loaded image: %p..%p\n",
+                node->base, node->base+node->n);
+#endif
+        last = node;
+        next = &(node->next);
+      }
+    else
+      {
+        _gst_mem.rememberedTableEntries--;
+        *next = node->next;
+        xfree (node);
+      }
 
+  _gst_mem.grey_pages.tail = last;
   _gst_fixup_object_pointers ();
 
   /* Now do the copying loop which will compact oldspace.  */
@@ -973,7 +1044,7 @@ _gst_compact (size_t new_heap_limit)
        oop < &_gst_mem.ot[_gst_mem.ot_size]; oop++)
     {
       PREFETCH_LOOP (oop, PREF_READ | PREF_NTA);
-      if ((oop->flags & (F_OLD | F_FIXED)) == F_OLD)
+      if ((oop->flags & (F_OLD | F_FIXED | F_LOADED)) == F_OLD)
         {
           gst_object new;
           size_t size = SIZE_TO_BYTES (TO_INT (oop->object->objSize));
@@ -1015,7 +1086,7 @@ _gst_global_gc (int next_allocation)
     
   if (!_gst_gc_running++
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity >= 2
       && !_gst_regression_testing)
     {
       /* print the first part of this message before we finish
@@ -1041,6 +1112,9 @@ _gst_global_gc (int next_allocation)
   _gst_mem.live_flags |= F_REACHABLE;
   check_weak_refs ();
   _gst_restore_object_pointers ();
+#if defined (GC_DEBUGGING)
+  _gst_check_oop_table ();
+#endif
   reset_incremental_gc (_gst_mem.ot);
 
   update_stats (&stats.timeOfLastGlobalGC,
@@ -1082,7 +1156,7 @@ _gst_global_gc (int next_allocation)
 
   if (!--_gst_gc_running
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity >= 2
       && !_gst_regression_testing)
     {
       fprintf (stderr, "%s\"\n", s);
@@ -1185,7 +1259,7 @@ _gst_finish_incremental_gc ()
 {
   OOP oop, firstOOP;
 
-#if defined (GC_DEBUGGING)
+#if defined (GC_DEBUG_OUTPUT)
   printf ("Completing sweep (%p...%p), validity flags %x\n", _gst_mem.last_swept_oop,
           _gst_mem.highest_swept_oop, _gst_mem.live_flags);
 #endif
@@ -1307,7 +1381,7 @@ reset_incremental_gc (OOP firstOOP)
     _gst_realloc_oop_table (_gst_mem.ot_size
       * (100 + _gst_mem.space_grow_rate) / 100);
 
-#if defined (GC_DEBUGGING)
+#if defined (GC_DEBUG_OUTPUT)
   printf ("Found unallocated at OOP %p, last allocated OOP %p\n"
           "Highest OOP swept top to bottom %p, lowest swept bottom to top %p\n",
 	  _gst_mem.first_allocated_oop, _gst_mem.last_allocated_oop,
@@ -1326,12 +1400,10 @@ _gst_sweep_oop (OOP oop)
 
 #ifdef ENABLE_JIT_TRANSLATION
   if (oop->flags & F_XLAT)
-    /* Unreachable, always free the native code.  Note
-       it is *not* optional to free the code in this case -- and
-       I'm not talking about memory leaks: a different
-       method could use the same OOP as this one and the
-       old method one would be executed instead of the new
-       one! */
+    /* Unreachable, always free the native code.  It is *not* optional
+       to free the code in this case -- and I'm not talking about memory
+       leaks: a different method could use the same OOP as this one and
+       the old method would be executed instead of the new one! */
     _gst_release_native_code (oop);
 #endif
 
@@ -1344,17 +1416,19 @@ _gst_sweep_oop (OOP oop)
       _gst_mem.numOldOOPs--;
       stats.reclaimedOldSpaceBytesSinceLastGlobalGC +=
 	SIZE_TO_BYTES (TO_INT (OOP_TO_OBJ (oop)->objSize));
-      _gst_mem_free (_gst_mem.fixed, oop->object);
+      if ((oop->flags & F_LOADED) == 0)
+        _gst_mem_free (_gst_mem.fixed, oop->object);
     }
   else if UNCOMMON (oop->flags & F_OLD)
     {
       _gst_mem.numOldOOPs--;
       stats.reclaimedOldSpaceBytesSinceLastGlobalGC +=
 	SIZE_TO_BYTES (TO_INT (OOP_TO_OBJ (oop)->objSize));
-      _gst_mem_free (_gst_mem.old, oop->object);
+      if ((oop->flags & F_LOADED) == 0)
+        _gst_mem_free (_gst_mem.old, oop->object);
     }
 
-  oop->flags = F_FREE;
+  oop->flags = 0;
 }
 
 void
@@ -1412,7 +1486,7 @@ queue_put (surv_space *q, OOP *src, int n)
       result = q->allocPtr;
       newAlloc = q->allocPtr + n;
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Top %p alloc %p tenure %p\n", q->topPtr, q->allocPtr, q->tenurePtr);
 #endif
 
@@ -1420,7 +1494,7 @@ queue_put (surv_space *q, OOP *src, int n)
         /* We tenure old objects as we copy more objects into
            the circular survivor space.  */
         {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
 	  printf ("Tenure: current max %p, needed %p\n", q->tenurePtr, newAlloc);
 #endif
 	  tenure_one_object();
@@ -1429,7 +1503,7 @@ queue_put (surv_space *q, OOP *src, int n)
 
       if UNCOMMON (newAlloc > q->maxPtr)
         {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
           printf ("Wrap: survivor space ends at %p, needed %p\n", q->maxPtr, newAlloc);
 #endif
           q->topPtr = q->allocPtr;
@@ -1461,14 +1535,14 @@ tenure_one_object ()
   OOP oop;
 
   oop = *_gst_mem.tenuring_queue.tenurePtr;
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("      ");
   _gst_display_oop (oop);
 #endif
 
   if (_gst_mem.scan.current == oop)
     {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Tenured OOP %p was being scanned\n", oop);
 #endif
 
@@ -1478,7 +1552,7 @@ tenure_one_object ()
 
   else if (_gst_mem.scan.queue_at == _gst_mem.tenuring_queue.tenurePtr)
     {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Tenured OOP %p had not been scanned yet\n", oop);
 #endif
 
@@ -1641,31 +1715,57 @@ copy_oops (void)
 }
 
 void
-scan_grey_pages ()
+_gst_print_grey_list (mst_Boolean check_pointers)
 {
-  grey_area_node *node, **next, *last;
+  grey_area_node *node;
   OOP *pOOP, oop;
   int i, n;
 
-#if defined (MMAN_DEBUGGING)
-  printf ("Pages on the grey list: ");
   for (n = 0, node = _gst_mem.grey_pages.head; node; node = node->next, n++)
     {
-      if (n & 3)
-        printf ("%12p ", node->base);
-      else
-        printf ("\n%12p ", node->base);
+      int new_pointers = 0;
+      if (check_pointers)
+        for (new_pointers = 0, pOOP = node->base, i = node->n; i--; pOOP++)
+          {
+            PREFETCH_LOOP (pOOP, PREF_READ | PREF_NTA);
+            oop = *pOOP;
+
+            /* Not all addresses are known to contain valid OOPs! */
+	    if (!IS_OOP_ADDR (oop))
+	      continue;
+
+            if (!IS_OOP_NEW (oop))
+	      continue;
+
+	    new_pointers++;
+	  }
+
+      printf ("%11p%c ", node->base, new_pointers == 0 ? ' ' : '*');
+      if ((n & 3) == 3)
+	putchar ('\n');
     }
 
   if (_gst_mem.grey_pages.tail)
     printf ("  (tail = %12p)", _gst_mem.grey_pages.tail->base);
 
   printf ("\n");
+}
+
+void
+scan_grey_pages ()
+{
+  grey_area_node *node, **next, *last;
+  OOP *pOOP, oop;
+  int i, n;
+
+#if defined (MMAN_DEBUG_OUTPUT)
+  printf ("Pages on the grey list:\n");
+  _gst_print_grey_list (true);
 #endif
 
   for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
     {
-#if defined(GC_DEBUGGING) || defined(MMAN_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT) || defined(MMAN_DEBUG_OUTPUT)
       printf ("Scanning grey page %p...%p\n", node->base, node->base + node->n);
 #endif
 
@@ -1692,7 +1792,7 @@ scan_grey_pages ()
 	{
           /* The entry was temporary, or we found no new-space
              pointers in it.  Delete it and make the page read-only.  */
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
 	  printf ("Protecting %p\n", node->base);
 #endif
 	  _gst_mem.rememberedTableEntries--;
@@ -1712,20 +1812,9 @@ scan_grey_pages ()
 
   _gst_mem.grey_pages.tail = last;
 
-#if defined (MMAN_DEBUGGING)
-  printf ("Pages left on the grey list: ");
-  for (n = 0, node = _gst_mem.grey_pages.head; node; node = node->next, n++)
-    {
-      if (n & 3)
-        printf ("%12p ", node->base);
-      else
-        printf ("\n%12p ", node->base);
-    }
-
-  if (_gst_mem.grey_pages.tail)
-    printf ("  (tail = %12p)", _gst_mem.grey_pages.tail->base);
-
-  printf ("\n");
+#if defined (MMAN_DEBUG_OUTPUT)
+  printf ("Pages left on the grey list:\n");
+  _gst_print_grey_list (false);
 #endif
 }
 
@@ -1745,7 +1834,7 @@ scan_grey_objects()
 	/* Objects might have moved, so update node->base.  */
 	node->base = (OOP *) &obj->data[1];
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Scanning grey range %p...%p (%p)\n", node->base, node->base + node->n, oop);
 #endif
 
@@ -1813,7 +1902,7 @@ scanned_fields_in (gst_object object,
 void
 cheney_scan (void)
 {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("Starting Cheney scan\n");
 #endif
 
@@ -1834,7 +1923,9 @@ cheney_scan (void)
 #if defined(GC_DEBUGGING)
       if (!IS_OOP_ADDR (oop))
 	abort();
+#endif
 
+#if defined(GC_DEBUG_OUTPUT)
       printf (">Scan ");
       _gst_display_oop (oop);
 #endif
@@ -1853,7 +1944,7 @@ cheney_scan (void)
 	MAYBE_COPY_OOP (_gst_mem.scan.at[i+1]);
     }
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("Ending Cheney scan\n");
 #endif
 }
@@ -1878,12 +1969,12 @@ _gst_copy_an_oop (OOP oop)
       obj = OOP_TO_OBJ (oop);
       pData = (OOP *) obj;
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf (">Copy ");
       _gst_display_oop (oop);
 #endif
 
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
       if UNCOMMON (!IS_INT (obj->objSize))
 	{
 	  printf ("Size not an integer in OOP %p (%p)\n", oop, obj);
@@ -1895,20 +1986,13 @@ _gst_copy_an_oop (OOP oop)
 	  printf ("Invalid size for OOP %p (%p)\n", oop, obj);
 	  abort ();
 	}
-#endif
 
-      if UNCOMMON (oop->flags & F_FREE)
+      if UNCOMMON (oop->flags == 0)
 	{
 	  printf ("Free OOP %p was referenced\n", oop);
-	  /* If we're lucky, we can print some information on it...
-	     if we're not, we segfault: not that bad given that
-	     we'll abort pretty soon! */
-	  oop->flags &= ~F_FREE;
-	  _gst_display_oop (oop);
 	  abort ();
 	}
 
-#if !defined (OPTIMIZE)
       if UNCOMMON ((oop->flags & F_OLD) ||
 	  IS_SURVIVOR_ADDR(obj, _gst_mem.active_half == &_gst_mem.surv[1]))
         {
@@ -1963,7 +2047,7 @@ mark_oops (void)
 void
 mark_ephemeron_oops (void)
 {
-  OOP *pOOP, *base;
+  OOP *pOOP, *pDeadOOP, *base;
   int i, size;
 
   /* Make a local copy of the buffer */
@@ -1987,9 +2071,9 @@ mark_ephemeron_oops (void)
       key->flags |= F_REACHABLE;
     }
 
-  for (pOOP = base, i = size; i--; pOOP++)
+  for (pOOP = pDeadOOP = base, i = size; i--; )
     {
-      OOP oop = *pOOP;
+      OOP oop = *pOOP++;
       gst_object obj = OOP_TO_OBJ(oop);
       OOP key = obj->data[0];
       int num = NUM_OOPS(obj);
@@ -2002,10 +2086,10 @@ mark_ephemeron_oops (void)
       for (j = 1; j < num; j++)
         MAYBE_MARK_OOP (obj->data[j]);
 
-      /* Remember that above we cleared F_EPHEMERON is the key
+      /* Remember that above we cleared F_EPHEMERON if the key
          is alive.  */
       if (!IS_OOP_MARKED (key) && (oop->flags & F_EPHEMERON))
-        _gst_add_buf_pointer (oop);
+        *pDeadOOP++ = oop;
 
       /* Ok, now mark the key.  */
       MAYBE_MARK_OOP (key);
@@ -2013,6 +2097,12 @@ mark_ephemeron_oops (void)
       /* Restore the flag in case it was cleared.  */
       oop->flags |= F_EPHEMERON;
     }
+
+  /* If more ephemerons were reachable from the object, go on...  */
+  if (_gst_buffer_size ())
+    mark_ephemeron_oops ();
+
+  _gst_add_buf_data (base, (char *) pDeadOOP - (char *) base);
 }
 
 #define TAIL_MARK_OOP(newOOP) BEGIN_MACRO { \
@@ -2038,7 +2128,7 @@ _gst_mark_an_oop_internal (OOP oop,
     {
       if (!oop)
 	{			/* in the loop! */
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
 	  gst_object obj = (gst_object) (curOOP - 1);	/* for debugging */
 #endif
 	iterationLoop:
@@ -2048,7 +2138,7 @@ _gst_mark_an_oop_internal (OOP oop,
 	  curOOP++;
 	  if (IS_OOP (oop))
 	    {
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
 	      if UNCOMMON (!IS_OOP_ADDR (oop))
 		{
 		  printf
@@ -2086,7 +2176,7 @@ _gst_mark_an_oop_internal (OOP oop,
 	  gst_object object;
 	  uintptr_t size;
 
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
 	  if UNCOMMON (IS_OOP_FREE (oop))
 	    {
 	      printf ("Error! Free OOP %p is being marked!\n", oop);
@@ -2095,7 +2185,7 @@ _gst_mark_an_oop_internal (OOP oop,
 	    }
 #endif
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
 	  printf (">Mark ");
 	  _gst_display_oop (oop);
 #endif
