@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later
  * version.
  *
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -914,15 +936,15 @@ _gst_compute_stack_positions (gst_uchar * bp,
 	    }
 
 	    SEND_ARITH {
-	      balance -= _gst_builtin_selectors[n]->numArgs;
+	      balance -= _gst_builtin_selectors[n].numArgs;
 	    }
 
 	    SEND_IMMEDIATE {
-	      balance -= super + _gst_builtin_selectors[n]->numArgs;
+	      balance -= super + _gst_builtin_selectors[n].numArgs;
 	    }
 
             SEND_SPECIAL {
-	      balance -= _gst_builtin_selectors[n + 16]->numArgs;
+	      balance -= _gst_builtin_selectors[n + 16].numArgs;
 	    }
 
             INVALID {
@@ -1370,10 +1392,10 @@ _gst_verify_method (OOP methodOOP, int *num_outer_temps, int depth)
 	    }
 
 	    SEND_ARITH {
-	      if (!_gst_builtin_selectors[n])
+	      if (!_gst_builtin_selectors[n].symbol)
 		return ("invalid immediate send");
 
-	      balance -= _gst_builtin_selectors[n]->numArgs;
+	      balance -= _gst_builtin_selectors[n].numArgs;
 
 	      /* Sends touch the new stack top, so they require an extra slot.  */
 	      if (curr_sp + balance < 1)
@@ -1381,10 +1403,10 @@ _gst_verify_method (OOP methodOOP, int *num_outer_temps, int depth)
 	    }
 
             SEND_SPECIAL {
-	      if (!_gst_builtin_selectors[n + 16])
+	      if (!_gst_builtin_selectors[n + 16].symbol)
 		return ("invalid immediate send");
 
-	      balance -= _gst_builtin_selectors[n + 16]->numArgs;
+	      balance -= _gst_builtin_selectors[n + 16].numArgs;
 
 	      /* Sends touch the new stack top, so they require an extra slot.  */
 	      if (curr_sp + balance < 1)
@@ -1392,10 +1414,10 @@ _gst_verify_method (OOP methodOOP, int *num_outer_temps, int depth)
 	    }
 
             SEND_IMMEDIATE {
-	      if (!_gst_builtin_selectors[n])
+	      if (!_gst_builtin_selectors[n].symbol)
 		return ("invalid immediate send");
 
-	      balance -= super + _gst_builtin_selectors[n]->numArgs;
+	      balance -= super + _gst_builtin_selectors[n].numArgs;
 
 	      /* Sends touch the new stack top, so they require an extra slot.  */
 	      if (curr_sp + balance < 1)
@@ -1704,11 +1726,11 @@ _gst_verify_method (OOP methodOOP, int *num_outer_temps, int depth)
 	    }
 
 	    SEND_ARITH {
-	      sp -= _gst_builtin_selectors[n]->numArgs;
+	      sp -= _gst_builtin_selectors[n].numArgs;
 	      sp[-1] = FROM_INT (VARYING);
 	    }
 	    SEND_SPECIAL {
-	      sp -= _gst_builtin_selectors[n + 16]->numArgs;
+	      sp -= _gst_builtin_selectors[n + 16].numArgs;
 	      sp[-1] = FROM_INT (VARYING);
 	    }
 
@@ -1747,7 +1769,7 @@ _gst_verify_method (OOP methodOOP, int *num_outer_temps, int depth)
 		    return (_gst_debug (), "Invalid send to super");
 
 		  last_used_literal = NULL;
-	          sp -= super + _gst_builtin_selectors[n]->numArgs;
+	          sp -= super + _gst_builtin_selectors[n].numArgs;
 	          if (super && sp[-1] != FROM_INT (SELF))
 		    return ("Invalid send to super");
 
