@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -72,72 +94,42 @@ typedef struct symbol_info
 symbol_info;
 
 
+/* These variables hold various symbols needed mostly by the compiler
+   and the C interface.  It is important that these symbols are *not*
+   included in the builtin selectors list (builtins.gperf) because
+   of the way we create symbols in _gst_init_symbols_pass1.  */
+
 OOP _gst_and_symbol = NULL;
-OOP _gst_as_scaled_decimal_scale_symbol = NULL;
-OOP _gst_at_put_symbol = NULL;
-OOP _gst_at_symbol = NULL;
-OOP _gst_at_end_symbol = NULL;
-OOP _gst_at_sign_symbol = NULL;
+OOP _gst_as_scaled_decimal_radix_scale_symbol = NULL;
 OOP _gst_bad_return_error_symbol = NULL;
-OOP _gst_bit_and_symbol = NULL;
-OOP _gst_bit_or_symbol = NULL;
-OOP _gst_bit_shift_symbol = NULL;
-OOP _gst_bit_xor_symbol = NULL;
 OOP _gst_boolean_symbol = NULL;
 OOP _gst_byte_array_out_symbol = NULL;
 OOP _gst_byte_array_symbol = NULL;
 OOP _gst_c_object_ptr_symbol = NULL;
 OOP _gst_c_object_symbol = NULL;
+OOP _gst_category_symbol = NULL;
 OOP _gst_char_symbol = NULL;
-OOP _gst_class_symbol = NULL;
-OOP _gst_divide_symbol = NULL;
-OOP _gst_do_symbol = NULL;
 OOP _gst_does_not_understand_symbol = NULL;
 OOP _gst_double_symbol = NULL;
-OOP _gst_equal_symbol = NULL;
 OOP _gst_false_symbol = NULL;
 OOP _gst_float_symbol = NULL;
-OOP _gst_greater_equal_symbol = NULL;
-OOP _gst_greater_than_symbol = NULL;
 OOP _gst_if_false_if_true_symbol = NULL;
 OOP _gst_if_false_symbol = NULL;
 OOP _gst_if_true_if_false_symbol = NULL;
 OOP _gst_if_true_symbol = NULL;
 OOP _gst_int_symbol = NULL;
-OOP _gst_integer_divide_symbol = NULL;
-OOP _gst_is_nil_symbol = NULL;
-OOP _gst_java_as_int_symbol = NULL;
-OOP _gst_java_as_long_symbol = NULL;
-OOP _gst_less_equal_symbol = NULL;
-OOP _gst_less_than_symbol = NULL;
 OOP _gst_long_double_symbol = NULL;
 OOP _gst_long_symbol = NULL;
-OOP _gst_minus_symbol = NULL;
 OOP _gst_must_be_boolean_symbol = NULL;
-OOP _gst_new_colon_symbol = NULL;
-OOP _gst_new_symbol = NULL;
-OOP _gst_next_put_symbol = NULL;
-OOP _gst_next_symbol = NULL;
 OOP _gst_nil_symbol = NULL;
-OOP _gst_not_equal_symbol = NULL;
-OOP _gst_not_nil_symbol = NULL;
-OOP _gst_not_same_object_symbol = NULL;
 OOP _gst_or_symbol = NULL;
-OOP _gst_perform_symbol = NULL;
-OOP _gst_perform_with_symbol = NULL;
-OOP _gst_perform_with_with_symbol = NULL;
-OOP _gst_perform_with_with_with_symbol = NULL;
-OOP _gst_perform_with_arguments_symbol = NULL;
 OOP _gst_permission_symbol = NULL;
-OOP _gst_plus_symbol = NULL;
 OOP _gst_primitive_symbol = NULL;
-OOP _gst_remainder_symbol = NULL;
 OOP _gst_repeat_symbol = NULL;
-OOP _gst_same_object_symbol = NULL;
 OOP _gst_self_smalltalk_symbol = NULL;
 OOP _gst_self_symbol = NULL;
-OOP _gst_size_symbol = NULL;
 OOP _gst_smalltalk_symbol = NULL;
+OOP _gst_smalltalk_namespace_symbol = NULL;
 OOP _gst_start_execution_symbol = NULL;
 OOP _gst_string_out_symbol = NULL;
 OOP _gst_string_symbol = NULL;
@@ -146,9 +138,7 @@ OOP _gst_symbol_symbol = NULL;
 OOP _gst_symbol_out_symbol = NULL;
 OOP _gst_symbol_table = NULL;
 OOP _gst_terminate_symbol = NULL;
-OOP _gst_this_context_symbol = NULL;
 OOP _gst_times_repeat_symbol = NULL;
-OOP _gst_times_symbol = NULL;
 OOP _gst_to_by_do_symbol = NULL;
 OOP _gst_to_do_symbol = NULL;
 OOP _gst_true_symbol = NULL;
@@ -156,11 +146,6 @@ OOP _gst_uint_symbol = NULL;
 OOP _gst_ulong_symbol = NULL;
 OOP _gst_undeclared_symbol = NULL;
 OOP _gst_unknown_symbol = NULL;
-OOP _gst_value_colon_symbol = NULL;
-OOP _gst_value_value_symbol = NULL;
-OOP _gst_value_value_value_symbol = NULL;
-OOP _gst_value_symbol = NULL;
-OOP _gst_value_with_arguments_symbol = NULL;
 OOP _gst_value_with_rec_with_args_symbol = NULL;
 OOP _gst_variadic_smalltalk_symbol = NULL;
 OOP _gst_variadic_symbol = NULL;
@@ -173,14 +158,15 @@ OOP _gst_while_false_colon_symbol = NULL;
 OOP _gst_while_false_symbol = NULL;
 OOP _gst_while_true_colon_symbol = NULL;
 OOP _gst_while_true_symbol = NULL;
-OOP _gst_yourself_symbol = NULL;
 OOP _gst_current_namespace = NULL;
 
+OOP temporaries_dictionary = NULL;
+
 /* The list of selectors for the send immediate bytecode.  */
-struct builtin_selector *_gst_builtin_selectors[256] = {};
+struct builtin_selector _gst_builtin_selectors[256] = {};
 
 /* True if undeclared globals can be considered forward references.  */
-int _gst_use_undeclared = 0;
+enum undeclared_strategy _gst_use_undeclared = UNDECLARED_TEMPORARIES;
 
 /* Answer whether OOP is a Smalltalk String LEN characters long and
    these characters match the first LEN characters of STR (which must
@@ -189,6 +175,15 @@ static mst_Boolean is_same_string (const char *str,
 				   OOP oop,
 				   int len);
 
+/* Allocate memory for a symbol of length LEN and whose contents are STR.
+   This function does not fill in the object's class because it is called
+   upon image loading, when the classes have not been initialized yet.  */
+static OOP alloc_symbol_oop (const char *str, int len);
+
+/* Link SYMBOLOOP into the symbol table (using the given hash table index),
+   and fill the class slot of the symbol.  */
+static OOP alloc_symlink (OOP symbolOOP, uintptr_t index);
+
 /* Answer whether C is considered a white space character in Smalltalk
    programs.  */
 static mst_Boolean is_white_space (char c);
@@ -228,6 +223,14 @@ static OOP scan_name (const char ** pp);
 static OOP intern_counted_string (const char *str,
 				  int len);
 
+/* This is a hack.  It is the same as _gst_intern_string except that,
+   if the given symbol is pointed to by PTESTOOP, we increment
+   PTESTOOP and return the old value.  This works and speeds up image
+   loading, because we are careful to create the same symbols in
+   _gst_init_symbols_passN and _gst_restore_symbols.  */
+static inline OOP
+intern_string_fast (const char *str, OOP *pTestOOP);
+
 /* This looks for SYMBOL among the instance variables that the current
    class declares, and returns the index of the variable if one is
    found.  */
@@ -256,72 +259,37 @@ static scope cur_scope = NULL;
    and is used to restore the global variables upon image load.  */
 static const symbol_info sym_info[] = {
   {&_gst_and_symbol, "and:"},
-  {&_gst_as_scaled_decimal_scale_symbol, "asScaledDecimal:scale:"},
-  {&_gst_at_put_symbol, "at:put:"},
-  {&_gst_at_symbol, "at:"},
-  {&_gst_at_end_symbol, "atEnd"},
-  {&_gst_at_sign_symbol, "@"},
+  {&_gst_as_scaled_decimal_radix_scale_symbol, "asScaledDecimal:radix:scale:"},
   {&_gst_bad_return_error_symbol, "badReturnError"},
-  {&_gst_bit_and_symbol, "bitAnd:"},
-  {&_gst_bit_or_symbol, "bitOr:"},
-  {&_gst_bit_xor_symbol, "bitXor:"},
-  {&_gst_bit_shift_symbol, "bitShift:"},
   {&_gst_byte_array_symbol, "byteArray"},
   {&_gst_byte_array_out_symbol, "byteArrayOut"},
   {&_gst_boolean_symbol, "boolean"},
-  {&_gst_char_symbol, "char"},
-  {&_gst_class_symbol, "class"},
   {&_gst_c_object_symbol, "cObject"},
   {&_gst_c_object_ptr_symbol, "cObjectPtr"},
-  {&_gst_divide_symbol, "/"},
-  {&_gst_do_symbol, "do:"},
+  {&_gst_category_symbol, "category:"},
+  {&_gst_char_symbol, "char"},
   {&_gst_does_not_understand_symbol, "doesNotUnderstand:"},
   {&_gst_float_symbol, "float"},
   {&_gst_double_symbol, "double"},
-  {&_gst_equal_symbol, "="},
   {&_gst_false_symbol, "false"},
-  {&_gst_greater_equal_symbol, ">="},
-  {&_gst_greater_than_symbol, ">"},
   {&_gst_if_false_if_true_symbol, "ifFalse:ifTrue:"},
   {&_gst_if_false_symbol, "ifFalse:"},
   {&_gst_if_true_if_false_symbol, "ifTrue:ifFalse:"},
   {&_gst_if_true_symbol, "ifTrue:"},
-  {&_gst_integer_divide_symbol, "//"},
   {&_gst_int_symbol, "int"},
   {&_gst_uint_symbol, "uInt"},
-  {&_gst_is_nil_symbol, "isNil"},
-  {&_gst_java_as_int_symbol, "javaAsInt"},
-  {&_gst_java_as_long_symbol, "javaAsLong"},
-  {&_gst_less_equal_symbol, "<="},
-  {&_gst_less_than_symbol, "<"},
   {&_gst_long_double_symbol, "longDouble"},
   {&_gst_long_symbol, "long"},
   {&_gst_ulong_symbol, "uLong"},
-  {&_gst_minus_symbol, "-"},
   {&_gst_must_be_boolean_symbol, "mustBeBoolean"},
-  {&_gst_new_colon_symbol, "new:"},
-  {&_gst_new_symbol, "new"},
-  {&_gst_next_put_symbol, "nextPut:"},
-  {&_gst_next_symbol, "next"},
   {&_gst_nil_symbol, "nil"},
-  {&_gst_not_equal_symbol, "~="},
-  {&_gst_not_nil_symbol, "notNil"},
-  {&_gst_not_same_object_symbol, "~~"},
   {&_gst_or_symbol, "or:"},
-  {&_gst_perform_symbol, "perform:"},
-  {&_gst_perform_with_symbol, "perform:with:"},
-  {&_gst_perform_with_with_symbol, "perform:with:with:"},
-  {&_gst_perform_with_with_with_symbol, "perform:with:with:with:"},
-  {&_gst_perform_with_arguments_symbol, "perform:withArguments:"},
-  {&_gst_plus_symbol, "+"},
   {&_gst_primitive_symbol, "primitive:"},
-  {&_gst_remainder_symbol, "\\\\"},
   {&_gst_repeat_symbol, "repeat"},
-  {&_gst_same_object_symbol, "=="},
   {&_gst_self_symbol, "self"},
   {&_gst_self_smalltalk_symbol, "selfSmalltalk"},
-  {&_gst_size_symbol, "size"},
   {&_gst_smalltalk_symbol, "smalltalk"},
+  {&_gst_smalltalk_namespace_symbol, "Smalltalk"},
   {&_gst_start_execution_symbol, "startExecution:"},
   {&_gst_string_out_symbol, "stringOut"},
   {&_gst_string_symbol, "string"},
@@ -329,22 +297,13 @@ static const symbol_info sym_info[] = {
   {&_gst_symbol_symbol, "symbol"},
   {&_gst_symbol_out_symbol, "symbolOut"},
   {&_gst_terminate_symbol, "__terminate"},
-  {&_gst_this_context_symbol, "thisContext"},
-  {&_gst_times_symbol, "*"},
   {&_gst_times_repeat_symbol, "timesRepeat:"},
   {&_gst_to_by_do_symbol, "to:by:do:"},
   {&_gst_to_do_symbol, "to:do:"},
   {&_gst_true_symbol, "true"},
   {&_gst_undeclared_symbol, "Undeclared"},
   {&_gst_unknown_symbol, "unknown"},
-  {&_gst_value_colon_symbol, "value:"},
-  {&_gst_value_value_symbol, "value:value:"},
-  {&_gst_value_value_value_symbol,
-   "value:value:value:"},
-  {&_gst_value_with_rec_with_args_symbol,
-   "valueWithReceiver:withArguments:"},
-  {&_gst_value_symbol, "value"},
-  {&_gst_value_with_arguments_symbol, "valueWithArguments:"},
+  {&_gst_value_with_rec_with_args_symbol, "valueWithReceiver:withArguments:"},
   {&_gst_variadic_symbol, "variadic"},
   {&_gst_variadic_smalltalk_symbol, "variadicSmalltalk"},
   {&_gst_vm_primitives_symbol, "VMPrimitives"},
@@ -356,7 +315,6 @@ static const symbol_info sym_info[] = {
   {&_gst_while_false_symbol, "whileFalse"},
   {&_gst_while_true_colon_symbol, "whileTrue:"},
   {&_gst_while_true_symbol, "whileTrue"},
-  {&_gst_yourself_symbol, "yourself"},
   {NULL, NULL},
 };
 
@@ -557,7 +515,7 @@ find_class_variable (OOP varName)
     {
       class = (gst_class) OOP_TO_OBJ (class_oop);
       assocOOP =
-	_gst_find_shared_pool_variable (class->environment, varName);
+	_gst_namespace_association_at (class->environment, varName);
       if (!IS_NIL (assocOOP))
 	return (assocOOP);
     }
@@ -572,7 +530,7 @@ find_class_variable (OOP varName)
 	{
 	  poolDictionaryOOP = ARRAY_AT (class->sharedPools, i + 1);
 	  assocOOP =
-	    _gst_find_shared_pool_variable (poolDictionaryOOP, varName);
+	    _gst_namespace_association_at (poolDictionaryOOP, varName);
 	  if (!IS_NIL (assocOOP))
 	    return (assocOOP);
 	}
@@ -582,10 +540,36 @@ find_class_variable (OOP varName)
 }
 
 
-OOP
-_gst_find_variable_binding (tree_node list)
+int
+_gst_set_undeclared (enum undeclared_strategy new)
 {
-  OOP symbol, root, assocOOP, undeclaredDictionary;
+  enum undeclared_strategy old = _gst_use_undeclared;
+  if (new != UNDECLARED_CURRENT)
+    _gst_use_undeclared = new;
+  return old;
+}
+
+OOP
+_gst_push_temporaries_dictionary (void)
+{
+  OOP old = temporaries_dictionary;
+  temporaries_dictionary = _gst_dictionary_new (8);
+  _gst_register_oop (temporaries_dictionary);
+  return old;
+}
+
+void
+_gst_pop_temporaries_dictionary (OOP dictionaryOOP)
+{
+  _gst_unregister_oop (temporaries_dictionary);
+  temporaries_dictionary = dictionaryOOP;
+}
+
+
+OOP
+_gst_find_variable_binding (tree_node list, mst_Boolean declare_temporary)
+{
+  OOP symbol, root, assocOOP, undeclaredDictionaryOOP;
 
   symbol = _gst_intern_string (list->v_list.name);
   assocOOP = find_class_variable (symbol);
@@ -594,7 +578,7 @@ _gst_find_variable_binding (tree_node list)
     {
       root = ASSOCIATION_VALUE (assocOOP);
       symbol = _gst_intern_string (list->v_list.name);
-      assocOOP = _gst_find_shared_pool_variable (root, symbol);
+      assocOOP = _gst_namespace_association_at (root, symbol);
     }
 
   if (IS_NIL (assocOOP) && !(list->v_list.next))
@@ -602,23 +586,27 @@ _gst_find_variable_binding (tree_node list)
       char *varName;
 
       varName = STRING_OOP_CHARS (symbol);
-      if (!isupper (*varName) || !_gst_use_undeclared)
-	return (assocOOP);
+      if (_gst_use_undeclared == UNDECLARED_TEMPORARIES
+	  && declare_temporary)
+        undeclaredDictionaryOOP = temporaries_dictionary;
 
-      undeclaredDictionary =
-	dictionary_at (_gst_smalltalk_dictionary,
-		       _gst_undeclared_symbol);
+      else if (_gst_use_undeclared == UNDECLARED_GLOBALS
+	       && isupper (*varName))
+        undeclaredDictionaryOOP = dictionary_at (_gst_smalltalk_dictionary,
+					         _gst_undeclared_symbol);
 
-      assocOOP =
-	dictionary_association_at (undeclaredDictionary, symbol);
+      else
+	undeclaredDictionaryOOP = _gst_nil_oop;
 
-      if (IS_NIL (assocOOP))
+      if (!IS_NIL (undeclaredDictionaryOOP))
 	{
-	  assocOOP =
-	    NAMESPACE_AT_PUT (undeclaredDictionary, symbol,
-			      _gst_nil_oop);
-
-	  MAKE_OOP_UNTRUSTED (assocOOP, _gst_untrusted_methods);
+          assocOOP = dictionary_association_at (undeclaredDictionaryOOP, symbol);
+          if (IS_NIL (assocOOP))
+	    {
+	      assocOOP =
+	        NAMESPACE_AT_PUT (undeclaredDictionaryOOP, symbol, _gst_nil_oop);
+	      MAKE_OOP_UNTRUSTED (assocOOP, _gst_untrusted_methods);
+	    }
 	}
     }
 
@@ -660,7 +648,7 @@ _gst_find_variable (symbol_entry * se,
       fill_symbol_entry (se, SCOPE_SPECIAL, true, symbol, NIL_INDEX, 0);
       return (true);
     }
-  else if (symbol == _gst_this_context_symbol)
+  else if (symbol == _gst_builtin_selectors[THIS_CONTEXT_SPECIAL].symbol)
     {
       fill_symbol_entry (se, SCOPE_SPECIAL, true, symbol,
 			 THIS_CONTEXT_INDEX, 0);
@@ -689,7 +677,7 @@ _gst_find_variable (symbol_entry * se,
       return (true);
     }
 
-  varAssoc = _gst_find_variable_binding (list);
+  varAssoc = _gst_find_variable_binding (list, true);
   if (IS_NIL (varAssoc))
     return (false);
 
@@ -1071,19 +1059,41 @@ _gst_intern_string (const char *str)
   return (intern_counted_string (str, len));
 }
 
+static uintptr_t
+hash_symbol (const char *str, int len)
+{
+  uintptr_t index = scramble (_gst_hash_string (str, len));
+  return (index & (SYMBOL_TABLE_SIZE - 1)) + 1;
+}
+
+static OOP
+alloc_symlink (OOP symbolOOP, uintptr_t index)
+{
+  gst_symbol symbol;
+  sym_link link;
+  OOP linkOOP;
+
+  symbol = (gst_symbol) OOP_TO_OBJ (symbolOOP);
+  symbol->objClass = _gst_symbol_class;
+
+  link = (sym_link) new_instance (_gst_sym_link_class, &linkOOP);
+  link->nextLink = ARRAY_AT (_gst_symbol_table, index);
+  link->symbol = symbolOOP;
+  ARRAY_AT_PUT (_gst_symbol_table, index, linkOOP);
+
+  return (symbolOOP);
+}
+
 static OOP
 intern_counted_string (const char *str,
 		       int len)
 {
   uintptr_t index;
-  sym_link link;
-  gst_symbol symbol;
   OOP symbolOOP, linkOOP;
+  sym_link link;
   inc_ptr incPtr;
 
-  index = scramble (_gst_hash_string (str, len));
-  index = (index & (SYMBOL_TABLE_SIZE - 1)) + 1;
-
+  index = hash_symbol (str, len);
   for (linkOOP = ARRAY_AT (_gst_symbol_table, index); !IS_NIL (linkOOP);
        linkOOP = link->nextLink)
     {
@@ -1098,22 +1108,31 @@ intern_counted_string (const char *str,
 #endif
 
   incPtr = INC_SAVE_POINTER ();
-  symbol = (gst_symbol) new_instance_with (_gst_symbol_class, 
-					   len, &symbolOOP);
-
-  memcpy (symbol->symString, str, len);
-  symbolOOP->flags |= F_READONLY;
+  symbolOOP = alloc_symbol_oop (str, len);
   INC_ADD_OOP (symbolOOP);
 
-  link = (sym_link) new_instance (_gst_sym_link_class, &linkOOP);
-  link->nextLink = ARRAY_AT (_gst_symbol_table, index);
-  link->symbol = symbolOOP;
-  ARRAY_AT_PUT (_gst_symbol_table, index, linkOOP);
-
+  alloc_symlink (symbolOOP, index);
   INC_RESTORE_POINTER (incPtr);
+
   return (symbolOOP);
 }
 
+static OOP
+alloc_symbol_oop (const char *str, int len)
+{
+  int numBytes, alignedBytes;
+  gst_symbol symbol;
+  OOP symbolOOP;
+
+  numBytes = sizeof(gst_object_header) + len;
+  alignedBytes = ROUNDED_BYTES (numBytes);
+  symbol = (gst_symbol) _gst_alloc_obj (alignedBytes, &symbolOOP);
+  INIT_UNALIGNED_OBJECT (symbolOOP, alignedBytes - numBytes);
+
+  memcpy (symbol->symString, str, len);
+  symbolOOP->flags |= F_READONLY;
+  return symbolOOP;
+}
 
 static mst_Boolean
 is_same_string (const char *str,
@@ -1237,25 +1256,22 @@ _gst_selector_num_args (OOP symbolOOP)
     return (1);
 
   for (numArgs = 0; len;)
-    {
-      if (bytes[--len] == ':')
-	{
-	  numArgs++;
-	}
-    }
+    if (bytes[--len] == ':')
+      numArgs++;
+
   return (numArgs);
 }
 
 #include "builtins.inl"
 
 void
-_gst_init_symbols (void)
+_gst_init_symbols_pass1 (void)
 {
   const symbol_info *si;
   struct builtin_selector *bs;
 
   for (si = sym_info; si->symbolVar; si++)
-    *si->symbolVar = _gst_intern_string (si->value);
+    *si->symbolVar = alloc_symbol_oop (si->value, strlen (si->value));
 
   /* Complete gperf's generated table with each symbol's OOP,
      and prepare a kind of reverse mapping from the 256 bytecodes
@@ -1267,7 +1283,70 @@ _gst_init_symbols (void)
     if (bs->offset != -1)
       {
 	const char *name = bs->offset + _gst_builtin_selectors_names;
-	bs->symbol = _gst_intern_string (name);
-        _gst_builtin_selectors[bs->bytecode] = bs;
+	bs->symbol = alloc_symbol_oop (name, strlen (name));
+        _gst_builtin_selectors[bs->bytecode] = *bs;
+      }
+}
+
+void
+_gst_init_symbols_pass2 (void)
+{
+  const symbol_info *si;
+  struct builtin_selector *bs;
+
+  for (si = sym_info; si->symbolVar; si++)
+    alloc_symlink (*si->symbolVar, hash_symbol (si->value, strlen (si->value)));
+
+  /* Complete gperf's generated table with each symbol's OOP,
+     and prepare a kind of reverse mapping from the 256 bytecodes
+     to the hash table entries.  */
+  for (bs = _gst_builtin_selectors_hash;
+       bs - _gst_builtin_selectors_hash <
+	 sizeof (_gst_builtin_selectors_hash) / sizeof (_gst_builtin_selectors_hash[0]);
+       bs++)
+    if (bs->offset != -1)
+      {
+	const char *name = bs->offset + _gst_builtin_selectors_names;
+	alloc_symlink (bs->symbol, hash_symbol (name, strlen (name)));
+      }
+}
+
+static inline OOP
+intern_string_fast (const char *str, OOP *pTestOOP)
+{
+  int len = strlen (str);
+  OOP testOOP = *pTestOOP;
+
+  if (is_same_string (str, testOOP, len))
+    {
+      (*pTestOOP)++;
+      return testOOP;
+    }
+  else
+    return intern_counted_string (str, len);
+}
+
+void
+_gst_restore_symbols (void)
+{
+  const symbol_info *si;
+  struct builtin_selector *bs;
+  OOP currentOOP = _gst_symbol_table + 1;
+
+  for (si = sym_info; si->symbolVar; si++)
+    *si->symbolVar = intern_string_fast (si->value, &currentOOP);
+
+  /* Complete gperf's generated table with each symbol's OOP,
+     and prepare a kind of reverse mapping from the 256 bytecodes
+     to the hash table entries.  */
+  for (bs = _gst_builtin_selectors_hash;
+       bs - _gst_builtin_selectors_hash <
+	 sizeof (_gst_builtin_selectors_hash) / sizeof (_gst_builtin_selectors_hash[0]);
+       bs++)
+    if (bs->offset != -1)
+      {
+	const char *name = bs->offset + _gst_builtin_selectors_names;
+	bs->symbol = intern_string_fast (name, &currentOOP);
+        _gst_builtin_selectors[bs->bytecode] = *bs;
       }
 }
